#
# Copyright (C) 2018-2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

from .. import _expression
from ._serializable import SerializableType, TypeParameterError
from ._primitive import UnsignedIntegerType, PrimitiveType, FloatType, ArithmeticType, IntegerType, BooleanType
from ._void import VoidType
from ._name import check_name, InvalidNameError


class InvalidConstantValueError(TypeParameterError):
    pass


class InvalidTypeError(TypeParameterError):
    pass


class Attribute(_expression.Any):
    def __init__(self, data_type: SerializableType, name: str):
        self._data_type = data_type
        self._name = str(name)

        if isinstance(data_type, VoidType):
            if self._name:
                raise InvalidNameError('Void-typed fields can be used only for padding and cannot be named')
        else:
            check_name(self._name)

    @property
    def data_type(self) -> SerializableType:
        return self._data_type

    @property
    def name(self) -> str:
        """For padding fields this is an empty string."""
        return self._name

    def __hash__(self) -> int:
        return hash((self._data_type, self._name))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Attribute):
            return (self._data_type == other._data_type) and (self._name == other.name)
        else:
            return NotImplemented  # pragma: no cover

    def __str__(self) -> str:
        """Returns the normalized DSDL representation of the attribute."""
        return ('%s %s' % (self.data_type, self.name)).strip()

    def __repr__(self) -> str:
        return '%s(data_type=%r, name=%r)' % (self.__class__.__name__, self.data_type, self.name)


class Field(Attribute):
    pass


class PaddingField(Field):
    def __init__(self, data_type: VoidType):
        if not isinstance(data_type, VoidType):
            raise TypeParameterError('Padding fields must be of the void type')

        super(PaddingField, self).__init__(data_type, '')


class Constant(Attribute):
    def __init__(self,
                 data_type: SerializableType,
                 name: str,
                 value: _expression.Any):
        super(Constant, self).__init__(data_type, name)

        if not isinstance(value, _expression.Primitive):
            raise InvalidConstantValueError('The constant value must be a primitive expression value')

        self._value = value
        del value

        # Interestingly, both the type of the constant and its value are instances of the same meta-type: expression.
        # BooleanType inherits from expression.Any, same as expression.Boolean.
        if isinstance(data_type, BooleanType):      # Boolean constant
            if not isinstance(self._value, _expression.Boolean):
                raise InvalidConstantValueError('Invalid value for boolean constant: %r' % self._value)

        elif isinstance(data_type, IntegerType):    # Integer constant
            if isinstance(self._value, _expression.Rational):
                if not self._value.is_integer():
                    raise InvalidConstantValueError('The value of an integer constant must be an integer; got %s' %
                                                    self._value)
            elif isinstance(self._value, _expression.String):
                as_bytes = self._value.native_value.encode('utf8')
                if len(as_bytes) != 1:
                    raise InvalidConstantValueError('A constant string must be exactly one ASCII character long')

                if not isinstance(data_type, UnsignedIntegerType) or data_type.bit_length != 8:
                    raise InvalidConstantValueError('Constant strings can be used only with uint8')

                self._value = _expression.Rational(ord(as_bytes))    # Replace string with integer
            else:
                raise InvalidConstantValueError('Invalid value type for integer constant: %r' % self._value)

        elif isinstance(data_type, FloatType):      # Floating point constant
            if not isinstance(self._value, _expression.Rational):
                raise InvalidConstantValueError('Invalid value type for float constant: %r' % self._value)

        else:
            raise InvalidTypeError('Invalid constant type: %r' % data_type)

        assert isinstance(self._value, _expression.Any)
        assert isinstance(self._value, _expression.Rational) == isinstance(self.data_type, (FloatType, IntegerType))
        assert isinstance(self._value, _expression.Boolean) == isinstance(self.data_type, BooleanType)

        # Range check
        if isinstance(self._value, _expression.Rational):
            assert isinstance(data_type, ArithmeticType)
            rng = data_type.inclusive_value_range
            if not (rng.min <= self._value.native_value <= rng.max):
                raise InvalidConstantValueError('Constant value %s exceeds the range of its data type %s' %
                                                (self._value, data_type))

    @property
    def value(self) -> _expression.Any:
        """
        The result of evaluating the constant initialization expression.
        The value is guaranteed to be compliant with the constant's own type -- it is checked at the evaluation time.
        The compliance rules are defined in the Specification.
        """
        return self._value

    def __hash__(self) -> int:
        return hash((self._data_type, self._name, self._value))

    def __eq__(self, other: object) -> bool:
        """Constants are equal if their type, name, and value are equal."""
        if isinstance(other, Constant):
            return super(Constant, self).__eq__(other) and (self._value == other._value)
        else:
            return NotImplemented  # pragma: no cover

    def __str__(self) -> str:
        """Returns the normalized DSDL representation of the constant and its value."""
        return '%s %s = %s' % (self.data_type, self.name, self.value)

    def __repr__(self) -> str:
        return 'Constant(data_type=%r, name=%r, value=%r)' % (self.data_type, self.name, self._value)


def _unittest_attribute() -> None:
    from pytest import raises
    from ._primitive import SignedIntegerType

    assert str(Field(BooleanType(PrimitiveType.CastMode.SATURATED), 'flag')) == 'saturated bool flag'
    assert repr(Field(BooleanType(PrimitiveType.CastMode.SATURATED), 'flag')) == \
        'Field(data_type=BooleanType(bit_length=1, cast_mode=<CastMode.SATURATED: 0>), name=\'flag\')'

    assert str(PaddingField(VoidType(32))) == 'void32'
    assert repr(PaddingField(VoidType(1))) == 'PaddingField(data_type=VoidType(bit_length=1), name=\'\')'

    assert Field(UnsignedIntegerType(1, PrimitiveType.CastMode.SATURATED), 'flag') == \
        Field(UnsignedIntegerType(1, PrimitiveType.CastMode.SATURATED), 'flag')
    assert hash(Field(UnsignedIntegerType(1, PrimitiveType.CastMode.SATURATED), 'flag')) == \
        hash(Field(UnsignedIntegerType(1, PrimitiveType.CastMode.SATURATED), 'flag'))

    assert Field(UnsignedIntegerType(1, PrimitiveType.CastMode.TRUNCATED), 'flag') != \
        Field(UnsignedIntegerType(1, PrimitiveType.CastMode.SATURATED), 'flag')
    assert hash(Field(UnsignedIntegerType(1, PrimitiveType.CastMode.TRUNCATED), 'flag')) != \
        hash(Field(UnsignedIntegerType(1, PrimitiveType.CastMode.SATURATED), 'flag'))

    with raises(TypeParameterError, match='.*void.*'):
        # noinspection PyTypeChecker
        repr(PaddingField(SignedIntegerType(8, PrimitiveType.CastMode.SATURATED)))   # type: ignore

    data_type = SignedIntegerType(32, PrimitiveType.CastMode.SATURATED)
    const = Constant(data_type, 'FOO_CONST', _expression.Rational(-123))
    assert str(const) == 'saturated int32 FOO_CONST = -123'
    assert const.data_type is data_type
    assert const.name == 'FOO_CONST'
    assert const.value == _expression.Rational(-123)

    assert repr(const) == 'Constant(data_type=%r, name=\'FOO_CONST\', value=rational(-123))' % data_type

    assert Constant(data_type, 'FOO_CONST', _expression.Rational(-123)) == const
    assert Constant(data_type, 'FOO_CONST', _expression.Rational(-124)) != const
    assert hash(Constant(data_type, 'FOO_CONST', _expression.Rational(-123))) == hash(const)
    assert hash(Constant(data_type, 'FOO_CONST', _expression.Rational(-124))) != hash(const)



"""Nodes that make up parse trees

Parsing spits out a tree of these, which you can then tell to walk itself and
spit out a useful value. Or you can walk it yourself; the structural attributes
are public.

"""
# TODO: If this is slow, think about using cElementTree or something.
from inspect import isfunction
from sys import version_info, exc_info

from six import reraise, python_2_unicode_compatible, with_metaclass, \
    iteritems

from parsimonious.exceptions import VisitationError, UndefinedLabel


@python_2_unicode_compatible
class Node(object):
    """A parse tree node

    Consider these immutable once constructed. As a side effect of a
    memory-saving strategy in the cache, multiple references to a single
    ``Node`` might be returned in a single parse tree. So, if you start
    messing with one, you'll see surprising parallel changes pop up elsewhere.

    My philosophy is that parse trees (and their nodes) should be
    representation-agnostic. That is, they shouldn't get all mixed up with what
    the final rendered form of a wiki page (or the intermediate representation
    of a programming language, or whatever) is going to be: you should be able
    to parse once and render several representations from the tree, one after
    another.

    """
    # I tried making this subclass list, but it got ugly. I had to construct
    # invalid ones and patch them up later, and there were other problems.
    __slots__ = ['expr',  # The expression that generated me
                 'full_text',  # The full text fed to the parser
                 'start', # The position in the text where that expr started matching
                 'end',   # The position after start where the expr first didn't
                          # match. [start:end] follow Python slice conventions.
                 'children']  # List of child parse tree nodes

    def __init__(self, expr, full_text, start, end, children=None):
        self.expr = expr
        self.full_text = full_text
        self.start = start
        self.end = end
        self.children = children or []

    @property
    def expr_name(self):
        # backwards compatibility
        return self.expr.name

    def __iter__(self):
        """Support looping over my children and doing tuple unpacks on me.

        It can be very handy to unpack nodes in arg lists; see
        :class:`PegVisitor` for an example.

        """
        return iter(self.children)

    @property
    def text(self):
        """Return the text this node matched."""
        return self.full_text[self.start:self.end]

    # From here down is just stuff for testing and debugging.

    def prettily(self, error=None):
        """Return a unicode, pretty-printed representation of me.

        :arg error: The node to highlight because an error occurred there

        """
        # TODO: If a Node appears multiple times in the tree, we'll point to
        # them all. Whoops.
        def indent(text):
            return '\n'.join(('    ' + line) for line in text.splitlines())
        ret = [u'<%s%s matching "%s">%s' % (
            self.__class__.__name__,
            (' called "%s"' % self.expr_name) if self.expr_name else '',
            self.text,
            '  <-- *** We were here. ***' if error is self else '')]
        for n in self:
            ret.append(indent(n.prettily(error=error)))
        return '\n'.join(ret)

    def __str__(self):
        """Return a compact, human-readable representation of me."""
        return self.prettily()

    def __eq__(self, other):
        """Support by-value deep comparison with other nodes for testing."""
        if not isinstance(other, Node):
            return NotImplemented

        return (self.expr == other.expr and
                self.full_text == other.full_text and
                self.start == other.start and
                self.end == other.end and
                self.children == other.children)

    def __ne__(self, other):
        return not self == other

    def __repr__(self, top_level=True):
        """Return a bit of code (though not an expression) that will recreate
        me."""
        # repr() of unicode flattens everything out to ASCII, so we don't need
        # to explicitly encode things afterward.
        ret = ["s = %r" % self.full_text] if top_level else []
        ret.append("%s(%r, s, %s, %s%s)" % (
            self.__class__.__name__,
            self.expr,
            self.start,
            self.end,
            (', children=[%s]' %
             ', '.join([c.__repr__(top_level=False) for c in self.children]))
            if self.children else ''))
        return '\n'.join(ret)


class RegexNode(Node):
    """Node returned from a ``Regex`` expression

    Grants access to the ``re.Match`` object, in case you want to access
    capturing groups, etc.

    """
    __slots__ = ['match']


class RuleDecoratorMeta(type):
    def __new__(metaclass, name, bases, namespace):
        def unvisit(name):
            """Remove any leading "visit_" from a method name."""
            return name[6:] if name.startswith('visit_') else name

        methods = [v for k, v in iteritems(namespace) if
                   hasattr(v, '_rule') and isfunction(v)]
        if methods:
            from parsimonious.grammar import Grammar  # circular import dodge

            methods.sort(key=(lambda x: x.func_code.co_firstlineno)
                             if version_info[0] < 3 else
                             (lambda x: x.__code__.co_firstlineno))
            # Possible enhancement: once we get the Grammar extensibility story
            # solidified, we can have @rules *add* to the default grammar
            # rather than pave over it.
            namespace['grammar'] = Grammar(
                '\n'.join('{name} = {expr}'.format(name=unvisit(m.__name__),
                                                   expr=m._rule)
                          for m in methods))
        return super(RuleDecoratorMeta,
                     metaclass).__new__(metaclass, name, bases, namespace)


class NodeVisitor(with_metaclass(RuleDecoratorMeta, object)):
    """A shell for writing things that turn parse trees into something useful

    Performs a depth-first traversal of an AST. Subclass this, add methods for
    each expr you care about, instantiate, and call
    ``visit(top_node_of_parse_tree)``. It'll return the useful stuff. This API
    is very similar to that of ``ast.NodeVisitor``.

    These could easily all be static methods, but that would add at least as
    much weirdness at the call site as the ``()`` for instantiation. And this
    way, we support subclasses that require state: options, for example, or a
    symbol table constructed from a programming language's AST.

    We never transform the parse tree in place, because...

    * There are likely multiple references to the same ``Node`` object in a
      parse tree, and changes to one reference would surprise you elsewhere.
    * It makes it impossible to report errors: you'd end up with the "error"
      arrow pointing someplace in a half-transformed mishmash of nodes--and
      that's assuming you're even transforming the tree into another tree.
      Heaven forbid you're making it into a string or something else.

    """

    #: The :term:`default grammar`: the one recommended for use with this
    #: visitor. If you populate this, you will be able to call
    #: :meth:`NodeVisitor.parse()` as a shortcut.
    grammar = None

    #: Classes of exceptions you actually intend to raise during visitation
    #: and which should propagate out of the visitor. These will not be
    #: wrapped in a VisitationError when they arise.
    unwrapped_exceptions = ()

    # TODO: If we need to optimize this, we can go back to putting subclasses
    # in charge of visiting children; they know when not to bother. Or we can
    # mark nodes as not descent-worthy in the grammar.
    def visit(self, node):
        """Walk a parse tree, transforming it into another representation.

        Recursively descend a parse tree, dispatching to the method named after
        the rule in the :class:`~parsimonious.grammar.Grammar` that produced
        each node. If, for example, a rule was... ::

            bold = '<b>'

        ...the ``visit_bold()`` method would be called. It is your
        responsibility to subclass :class:`NodeVisitor` and implement those
        methods.

        """
        method = getattr(self, 'visit_' + node.expr_name, self.generic_visit)

        # Call that method, and show where in the tree it failed if it blows
        # up.
        try:
            return method(node, [self.visit(n) for n in node])
        except (VisitationError, UndefinedLabel):
            # Don't catch and re-wrap already-wrapped exceptions.
            raise
        except self.unwrapped_exceptions:
            raise
        except Exception:
            # Catch any exception, and tack on a parse tree so it's easier to
            # see where it went wrong.
            exc_class, exc, tb = exc_info()
            reraise(VisitationError, VisitationError(exc, exc_class, node), tb)

    def generic_visit(self, node, visited_children):
        """Default visitor method

        :arg node: The node we're visiting
        :arg visited_children: The results of visiting the children of that
            node, in a list

        I'm not sure there's an implementation of this that makes sense across
        all (or even most) use cases, so we leave it to subclasses to implement
        for now.

        """
        raise NotImplementedError('No visitor method was defined for this expression: %s' %
                                  node.expr.as_rule())

    # Convenience methods:

    def parse(self, text, pos=0):
        """Parse some text with this Visitor's default grammar and return the
        result of visiting it.

        ``SomeVisitor().parse('some_string')`` is a shortcut for
        ``SomeVisitor().visit(some_grammar.parse('some_string'))``.

        """
        return self._parse_or_match(text, pos, 'parse')

    def match(self, text, pos=0):
        """Parse and visit some text with this Visitor's default grammar, but
        don't insist on parsing all the way to the end.

        ``SomeVisitor().match('some_string')`` is a shortcut for
        ``SomeVisitor().visit(some_grammar.match('some_string'))``.

        """
        return self._parse_or_match(text, pos, 'match')

    # Internal convenience methods to help you write your own visitors:

    def lift_child(self, node, children):
        """Lift the sole child of ``node`` up to replace the node."""
        first_child, = children
        return first_child

    # Private methods:

    def _parse_or_match(self, text, pos, method_name):
        """Execute a parse or match on the default grammar, followed by a
        visitation.

        Raise RuntimeError if there is no default grammar specified.

        """
        if not self.grammar:
            raise RuntimeError(
                "The {cls}.{method}() shortcut won't work because {cls} was "
                "never associated with a specific " "grammar. Fill out its "
                "`grammar` attribute, and try again.".format(
                    cls=self.__class__.__name__,
                    method=method_name))
        return self.visit(getattr(self.grammar, method_name)(text, pos=pos))


def rule(rule_string):
    """Decorate a NodeVisitor ``visit_*`` method to tie a grammar rule to it.

    The following will arrange for the ``visit_digit`` method to receive the
    results of the ``~"[0-9]"`` parse rule::

        @rule('~"[0-9]"')
        def visit_digit(self, node, visited_children):
            ...

    Notice that there is no "digit = " as part of the rule; that gets inferred
    from the method name.

    In cases where there is only one kind of visitor interested in a grammar,
    using ``@rule`` saves you having to look back and forth between the visitor
    and the grammar definition.

    On an implementation level, all ``@rule`` rules get stitched together into
    a :class:`~parsimonious.Grammar` that becomes the NodeVisitor's
    :term:`default grammar`.

    Typically, the choice of a default rule for this grammar is simple: whatever
    ``@rule`` comes first in the class is the default. But the choice may become
    surprising if you divide the ``@rule`` calls among subclasses. At the
    moment, which method "comes first" is decided simply by comparing line
    numbers, so whatever method is on the smallest-numbered line will be the
    default. In a future release, this will change to pick the
    first ``@rule`` call on the basemost class that has one. That way, a
    subclass which does not override the default rule's ``visit_*`` method
    won't unintentionally change which rule is the default.

    """
    def decorator(method):
        method._rule = rule_string  # XXX: Maybe register them on a class var instead so we can just override a @rule'd visitor method on a subclass without blowing away the rule string that comes with it.
        return method
    return decorator



#
# Copyright (C) 2018-2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import abc
import typing
import operator
import fractions
import unicodedata
from . import _any


# noinspection PyAbstractClass
class Primitive(_any.Any):
    @property
    @abc.abstractmethod
    def native_value(self) -> typing.Any:
        """
        Yields an appropriate Python-native representation of the contained value,
        like :class:`fractions.Fraction`, :class:`str`, etc.
        Specializations define covariant return types.
        """
        raise NotImplementedError  # pragma: no cover


class Boolean(Primitive):
    TYPE_NAME = 'bool'

    def __init__(self, value: bool = False):
        if not isinstance(value, bool):
            raise ValueError('Cannot construct a Boolean instance from ' + type(value).__name__)

        self._value = value  # type: bool

    @property
    def native_value(self) -> bool:
        return self._value

    def __hash__(self) -> int:
        return int(self._value)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Boolean):
            return self._value == other._value
        else:  # pragma: no cover
            return NotImplemented

    def __str__(self) -> str:
        return 'true' if self._value else 'false'

    def __bool__(self) -> bool:     # For use in expressions without accessing "native_value"
        return self._value

    def _logical_not(self) -> 'Boolean':
        return Boolean(not self._value)

    def _logical_and(self, right: _any.Any) -> 'Boolean':
        if isinstance(right, Boolean):
            return Boolean(self._value and right._value)
        else:
            raise _any.UndefinedOperatorError

    def _logical_or(self, right: _any.Any) -> 'Boolean':
        if isinstance(right, Boolean):
            return Boolean(self._value or right._value)
        else:
            raise _any.UndefinedOperatorError

    def _equal(self, right: _any.Any) -> 'Boolean':
        if isinstance(right, Boolean):
            return Boolean(self._value == right._value)
        else:
            raise _any.UndefinedOperatorError


class Rational(Primitive):
    TYPE_NAME = 'rational'

    def __init__(self, value: typing.Union[int, fractions.Fraction]):
        # We must support float as well, because some operators on Fraction sometimes yield float, e.g. power.
        if not isinstance(value, (int, float, fractions.Fraction)):
            raise ValueError('Cannot construct a Rational instance from ' + type(value).__name__)
        self._value = fractions.Fraction(value)  # type: fractions.Fraction

    @property
    def native_value(self) -> fractions.Fraction:
        return self._value

    def as_native_integer(self) -> int:
        """
        Returns the inferior as a native integer,
        unless it cannot be represented as such without the loss of precision; i.e., if denominator != 1,
        in which case an invalid operand exception is thrown.
        """
        if self.is_integer():
            return self._value.numerator
        else:
            raise _any.InvalidOperandError('Rational %s is not an integer' % self._value)

    def is_integer(self) -> bool:
        """Whether the demonimator equals one."""
        return self._value.denominator == 1

    def __hash__(self) -> int:
        return hash(self._value)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Rational):
            return self._value == other._value
        else:  # pragma: no cover
            return NotImplemented

    def __str__(self) -> str:
        return str(self._value)

    #
    # Unary operators.
    #
    def _positive(self) -> 'Rational':
        return Rational(+self._value)

    def _negative(self) -> 'Rational':
        return Rational(-self._value)

    #
    # Binary comparison operators.
    #
    def _generic_compare(self, right: _any.Any, impl: typing.Callable[[typing.Any, typing.Any], bool]) -> Boolean:
        if isinstance(right, Rational):
            return Boolean(impl(self._value, right._value))
        else:
            raise _any.UndefinedOperatorError

    def _equal(self, right: _any.Any) -> 'Boolean':
        return self._generic_compare(right, operator.eq)

    def _less_or_equal(self, right: _any.Any) -> 'Boolean':
        return self._generic_compare(right, operator.le)

    def _greater_or_equal(self, right: _any.Any) -> 'Boolean':
        return self._generic_compare(right, operator.ge)

    def _less(self, right: _any.Any) -> 'Boolean':
        return self._generic_compare(right, operator.lt)

    def _greater(self, right: _any.Any) -> 'Boolean':
        return self._generic_compare(right, operator.gt)

    #
    # Binary bitwise operators.
    #
    def _generic_bitwise(self, right: _any.Any, impl: typing.Callable[[typing.Any, typing.Any], typing.Any]) \
            -> 'Rational':
        if isinstance(right, Rational):
            return Rational(impl(self.as_native_integer(), right.as_native_integer()))    # Throws if not an integer.
        else:
            raise _any.UndefinedOperatorError

    def _bitwise_or(self, right: _any.Any) -> 'Rational':
        return self._generic_bitwise(right, operator.or_)

    def _bitwise_xor(self, right: _any.Any) -> 'Rational':
        return self._generic_bitwise(right, operator.xor)

    def _bitwise_and(self, right: _any.Any) -> 'Rational':
        return self._generic_bitwise(right, operator.and_)

    #
    # Binary arithmetic operators.
    #
    def _generic_arithmetic(self,
                            right: _any.Any,
                            impl: typing.Callable[[typing.Any, typing.Any], typing.Any]) -> 'Rational':
        if isinstance(right, Rational):
            try:
                result = impl(self._value, right._value)
            except ZeroDivisionError:
                raise _any.InvalidOperandError('Cannot divide %s by zero' % self._value)
            else:
                return Rational(result)
        else:
            raise _any.UndefinedOperatorError

    def _add(self, right: _any.Any) -> 'Rational':
        return self._generic_arithmetic(right, operator.add)

    def _subtract(self, right: _any.Any) -> 'Rational':
        return self._generic_arithmetic(right, operator.sub)

    def _multiply(self, right: _any.Any) -> 'Rational':
        return self._generic_arithmetic(right, operator.mul)

    def _divide(self, right: _any.Any) -> 'Rational':
        return self._generic_arithmetic(right, operator.truediv)

    def _modulo(self, right: _any.Any) -> 'Rational':
        return self._generic_arithmetic(right, operator.mod)

    def _power(self, right: _any.Any) -> 'Rational':
        return self._generic_arithmetic(right, operator.pow)


class String(Primitive):
    TYPE_NAME = 'string'

    def __init__(self, value: str):
        if not isinstance(value, str):
            raise ValueError('Cannot construct a String instance from ' + type(value).__name__)
        self._value = value  # type: str

    @property
    def native_value(self) -> str:
        return self._value

    def __hash__(self) -> int:
        return hash(self._value)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, String):
            return self._value == other._value
        else:  # pragma: no cover
            return NotImplemented

    def __str__(self) -> str:
        return repr(self._value)

    def _add(self, right: _any.Any) -> 'String':
        if isinstance(right, String):
            return String(self._value + right._value)
        else:
            raise _any.UndefinedOperatorError

    def _equal(self, right: _any.Any) -> Boolean:
        if isinstance(right, String):
            def normalized(s: str) -> str:
                return unicodedata.normalize('NFC', s)

            return Boolean(normalized(self._value) == normalized(right._value))
        else:
            raise _any.UndefinedOperatorError



#
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import os
import time
import typing
import logging
from . import _error
from . import _serializable
from . import _parser


_logger = logging.getLogger(__name__)


class FileNameFormatError(_error.InvalidDefinitionError):
    """
    Raised when a DSDL definition file is named incorrectly.
    """
    def __init__(self, text: str, path: str):
        super(FileNameFormatError, self).__init__(text=text, path=str(path))


class DSDLDefinition:
    """
    A DSDL type definition source abstracts the filesystem level details away, presenting a higher-level
    interface that operates solely on the level of type names, namespaces, fixed identifiers, and so on.
    Upper layers that operate on top of this abstraction do not concern themselves with the file system at all.
    """

    def __init__(self,
                 file_path: str,
                 root_namespace_path: str):
        # Normalizing the path and reading the definition text
        self._file_path = os.path.abspath(file_path)
        del file_path
        self._root_namespace_path = os.path.abspath(root_namespace_path)
        del root_namespace_path
        with open(self._file_path) as f:
            self._text = str(f.read())

        # Checking the sanity of the root directory path - can't contain separators
        if _serializable.CompositeType.NAME_COMPONENT_SEPARATOR in os.path.split(self._root_namespace_path)[-1]:
            raise FileNameFormatError('Invalid namespace name', path=self._root_namespace_path)

        # Determining the relative path within the root namespace directory
        relative_path = str(os.path.join(os.path.split(self._root_namespace_path)[-1],
                                         os.path.relpath(self._file_path, self._root_namespace_path)))

        relative_directory, basename = [str(x) for x in os.path.split(relative_path)]   # type: str, str

        # Parsing the basename, e.g., 434.GetTransportStatistics.0.1.uavcan
        basename_components = basename.split('.')[:-1]
        str_fixed_port_id = None    # type: typing.Optional[str]
        if len(basename_components) == 4:
            str_fixed_port_id, short_name, str_major_version, str_minor_version = basename_components
        elif len(basename_components) == 3:
            short_name, str_major_version, str_minor_version = basename_components
        else:
            raise FileNameFormatError('Invalid file name', path=self._file_path)

        # Parsing the fixed port ID, if specified; None if not
        if str_fixed_port_id is not None:
            try:
                self._fixed_port_id = int(str_fixed_port_id)    # type: typing.Optional[int]
            except ValueError:
                raise FileNameFormatError(
                    'Not a valid fixed port-ID: %r. '
                    'Namespaces are defined as directories; putting the namespace name in the file name will not work. '
                    'For example: "foo/Bar.1.0.uavcan" is OK (where "foo" is a directory); "foo.Bar.1.0.uavcan" is not.'
                    % str_fixed_port_id,
                    path=self._file_path,
                ) from None
        else:
            self._fixed_port_id = None

        # Parsing the version numbers
        try:
            self._version = _serializable.Version(major=int(str_major_version),
                                                  minor=int(str_minor_version))
        except ValueError:
            raise FileNameFormatError('Could not parse the version numbers', path=self._file_path) from None

        # Finally, constructing the name
        namespace_components = list(relative_directory.strip(os.sep).split(os.sep))
        for nc in namespace_components:
            if _serializable.CompositeType.NAME_COMPONENT_SEPARATOR in nc:
                raise FileNameFormatError('Invalid name for namespace component', path=self._file_path)

        self._name = _serializable.CompositeType.NAME_COMPONENT_SEPARATOR\
            .join(namespace_components + [str(short_name)])  # type: str

        self._cached_type = None    # type: typing.Optional[_serializable.CompositeType]

    def read(self,
             lookup_definitions:              typing.Iterable['DSDLDefinition'],
             print_output_handler:            typing.Callable[[int, str], None],
             allow_unregulated_fixed_port_id: bool) -> _serializable.CompositeType:
        """
        Reads the data type definition and returns its high-level data type representation.
        The output is cached; all following invocations will read from the cache.
        Caching is very important, because it is expected that the same definition may be referred to multiple
        times (e.g., for composition or when accessing external constants). Re-processing a definition every time
        it is accessed would be a huge waste of time.
        Note, however, that this may lead to unexpected complications if one is attempting to re-read a definition
        with different inputs (e.g., different lookup paths) expecting to get a different result: caching would
        get in the way. That issue is easy to avoid by creating a new instance of the object.
        :param lookup_definitions:              List of definitions available for referring to.
        :param print_output_handler:            Used for @print and for diagnostics: (line_number, text) -> None.
        :param allow_unregulated_fixed_port_id: Do not complain about fixed unregulated port IDs.
        :return: The data type representation.
        """
        log_prefix = '%s.%d.%d' % (self.full_name, self.version.major, self.version.minor)
        if self._cached_type is not None:
            _logger.debug('%s: Cache hit', log_prefix)
            return self._cached_type

        started_at = time.monotonic()

        # Remove the target definition from the lookup list in order to prevent
        # infinite recursion on self-referential definitions.
        lookup_definitions = list(filter(lambda d: d != self, lookup_definitions))

        _logger.info('%s: Starting processing with %d lookup definitions located in root namespaces: %s',
                     log_prefix, len(lookup_definitions),
                     ', '.join(set(sorted(map(lambda x: x.root_namespace, lookup_definitions)))))
        try:
            builder = _data_type_builder.DataTypeBuilder(
                definition=self,
                lookup_definitions=lookup_definitions,
                print_output_handler=print_output_handler,
                allow_unregulated_fixed_port_id=allow_unregulated_fixed_port_id
            )
            with open(self.file_path) as f:
                _parser.parse(f.read(), builder)

            self._cached_type = builder.finalize()

            _logger.info('%s: Processed in %.0f ms; category: %s, fixed port ID: %s',
                         log_prefix,
                         (time.monotonic() - started_at) * 1e3,
                         type(self._cached_type).__name__,
                         self._cached_type.fixed_port_id)
            return self._cached_type
        except _error.FrontendError as ex:                              # pragma: no cover
            ex.set_error_location_if_unknown(path=self.file_path)
            raise ex
        except Exception as ex:                                         # pragma: no cover
            raise _error.InternalError(culprit=ex, path=self.file_path)

    @property
    def full_name(self) -> str:
        """The full name, e.g., uavcan.node.Heartbeat"""
        return self._name

    @property
    def name_components(self) -> typing.List[str]:
        """Components of the full name as a list, e.g., ['uavcan', 'node', 'Heartbeat']"""
        return self._name.split(_serializable.CompositeType.NAME_COMPONENT_SEPARATOR)

    @property
    def short_name(self) -> str:
        """The last component of the full name, e.g., Heartbeat of uavcan.node.Heartbeat"""
        return self.name_components[-1]

    @property
    def full_namespace(self) -> str:
        """The full name without the short name, e.g., uavcan.node for uavcan.node.Heartbeat"""
        return str(_serializable.CompositeType.NAME_COMPONENT_SEPARATOR.join(self.name_components[:-1]))

    @property
    def root_namespace(self) -> str:
        """The first component of the full name, e.g., uavcan of uavcan.node.Heartbeat"""
        return self.name_components[0]

    @property
    def text(self) -> str:
        """The source text in its raw unprocessed form (with comments, formatting intact, and everything)"""
        return self._text

    @property
    def version(self) -> _serializable.Version:
        return self._version

    @property
    def fixed_port_id(self) -> typing.Optional[int]:
        """Either the fixed port ID as integer, or None if not defined for this type."""
        return self._fixed_port_id

    @property
    def has_fixed_port_id(self) -> bool:
        return self.fixed_port_id is not None

    @property
    def file_path(self) -> str:
        return self._file_path

    @property
    def root_namespace_path(self) -> str:
        return self._root_namespace_path

    def __eq__(self, other: object) -> bool:
        """
        Two definitions will compare equal if they share the same name AND version number.
        Definitions of the same name but different versions are not considered equal.
        """
        if isinstance(other, DSDLDefinition):
            return self.full_name == other.full_name and self.version == other.version
        else:  # pragma: no cover
            return NotImplemented

    def __str__(self) -> str:
        return 'DSDLDefinition(full_name=%r, version=%r, fixed_port_id=%r, file_path=%r)' % \
            (self.full_name, self.version, self.fixed_port_id, self.file_path)

    __repr__ = __str__


# Moved this import here to break recursive dependency.
# Maybe I have messed up the architecture? Should think about it later.
from . import _data_type_builder  # noqa: E402



#
# Copyright (C) 2018-2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import os
import typing
import logging
from . import _serializable
from . import _expression
from . import _error
from . import _dsdl_definition
from . import _parser
from . import _data_schema_builder
from . import _port_id_ranges


class AssertionCheckFailureError(_error.InvalidDefinitionError):
    pass


class UndefinedDataTypeError(_error.InvalidDefinitionError):
    pass


class UndefinedIdentifierError(_error.InvalidDefinitionError):
    pass


class InvalidDirectiveError(_error.InvalidDefinitionError):
    pass


class UnregulatedFixedPortIDError(_error.InvalidDefinitionError):
    pass


_logger = logging.getLogger(__name__)


class DataTypeBuilder(_parser.StatementStreamProcessor):
    def __init__(self,
                 definition:                      _dsdl_definition.DSDLDefinition,
                 lookup_definitions:              typing.Iterable[_dsdl_definition.DSDLDefinition],
                 print_output_handler:            typing.Callable[[int, str], None],
                 allow_unregulated_fixed_port_id: bool):
        self._definition = definition
        self._lookup_definitions = list(lookup_definitions)
        self._print_output_handler = print_output_handler
        self._allow_unregulated_fixed_port_id = allow_unregulated_fixed_port_id

        assert isinstance(self._definition, _dsdl_definition.DSDLDefinition)
        assert all(map(lambda x: isinstance(x, _dsdl_definition.DSDLDefinition), lookup_definitions))
        assert callable(self._print_output_handler)
        assert isinstance(self._allow_unregulated_fixed_port_id, bool)

        self._structs = [_data_schema_builder.DataSchemaBuilder()]
        self._is_deprecated = False

    def finalize(self) -> _serializable.CompositeType:
        if len(self._structs) == 1:     # Message type
            struct, = self._structs     # type: _data_schema_builder.DataSchemaBuilder,
            ty = _serializable.UnionType if struct.union else _serializable.StructureType
            fin = ty(name=self._definition.full_name,
                     version=self._definition.version,
                     attributes=struct.attributes,
                     deprecated=self._is_deprecated,
                     fixed_port_id=self._definition.fixed_port_id,
                     source_file_path=self._definition.file_path)  # type: _serializable.CompositeType
            if not struct.sealed:
                out = _serializable.DelimitedType(fin, extent=struct.extent)  # type: _serializable.CompositeType
                _logger.debug('%r wrapped into %r', fin, out)
                if struct.extent is None:
                    assert isinstance(out, _serializable.DelimitedType)
                    self._print_output_handler(1, self._render_explicit_extent_recommendation(out))
            else:
                assert struct.extent is None, 'Internal constraint violation'
                out = fin

        else:  # Service type
            request, response = self._structs
            assert isinstance(request, _data_schema_builder.DataSchemaBuilder)
            assert isinstance(response, _data_schema_builder.DataSchemaBuilder)
            # noinspection SpellCheckingInspection
            out = _serializable.ServiceType(
                name=self._definition.full_name,                        # pozabito vse na svete
                version=self._definition.version,                       # serdce zamerlo v grudi
                request_params=request.to_service_schema_params(),      # tolko nebo tolko veter
                response_params=response.to_service_schema_params(),    # tolko radost vperedi
                deprecated=self._is_deprecated,
                fixed_port_id=self._definition.fixed_port_id,
                source_file_path=self._definition.file_path,
            )
            assert isinstance(out, _serializable.ServiceType)
            req_ty, resp_ty = out.request_type, out.response_type
            if isinstance(req_ty, _serializable.DelimitedType) and request.extent is None:
                self._print_output_handler(1, 'Request: ' + self._render_explicit_extent_recommendation(req_ty))
            if isinstance(resp_ty, _serializable.DelimitedType) and response.extent is None:
                self._print_output_handler(1, 'Response: ' + self._render_explicit_extent_recommendation(resp_ty))

        assert isinstance(out, _serializable.CompositeType)
        if not self._allow_unregulated_fixed_port_id:
            port_id = out.fixed_port_id
            if port_id is not None:
                is_service_type = isinstance(out, _serializable.ServiceType)
                f = _port_id_ranges.is_valid_regulated_service_id if is_service_type else \
                    _port_id_ranges.is_valid_regulated_subject_id
                if not f(port_id, out.root_namespace):
                    raise UnregulatedFixedPortIDError(
                        'Regulated port ID %r for %s type %r is not valid. '
                        'Consider using allow_unregulated_fixed_port_id.' %
                        (port_id, 'service' if is_service_type else 'message', out.full_name))
        return out

    def on_constant(self,
                    constant_type: _serializable.SerializableType,
                    name: str,
                    value: _expression.Any) -> None:
        self._on_attribute()
        self._structs[-1].add_constant(_serializable.Constant(constant_type, name, value))

    def on_field(self, field_type: _serializable.SerializableType, name: str) -> None:
        self._on_attribute()
        self._structs[-1].add_field(_serializable.Field(field_type, name))

    def on_padding_field(self, padding_field_type: _serializable.VoidType) -> None:
        self._on_attribute()
        self._structs[-1].add_field(_serializable.PaddingField(padding_field_type))

    def on_directive(self,
                     line_number: int,
                     directive_name: str,
                     associated_expression_value: typing.Optional[_expression.Any]) -> None:
        try:
            handler = {
                'print':      self._on_print_directive,
                'assert':     self._on_assert_directive,
                'extent':     self._on_extent_directive,
                'sealed':     self._on_sealed_directive,
                'union':      self._on_union_directive,
                'deprecated': self._on_deprecated_directive,
            }[directive_name]
        except KeyError:
            raise InvalidDirectiveError('Unknown directive %r' % directive_name)
        else:
            assert callable(handler)
            return handler(line_number, associated_expression_value)

    def on_service_response_marker(self) -> None:
        if len(self._structs) > 1:
            raise _error.InvalidDefinitionError('Duplicated service response marker')

        self._structs.append(_data_schema_builder.DataSchemaBuilder())
        assert len(self._structs) == 2

    def resolve_top_level_identifier(self, name: str) -> _expression.Any:
        # Look only in the current data structure. The lookup cannot cross the service request/response boundary.
        for c in self._structs[-1].constants:
            if c.name == name:
                return c.value

        if name == '_offset_':
            bls = self._structs[-1].offset
            assert len(bls) > 0 and all(map(lambda x: isinstance(x, int), bls))
            return _expression.Set(map(_expression.Rational, bls))
        else:
            raise UndefinedIdentifierError('Undefined identifier: %r' % name)

    def resolve_versioned_data_type(self, name: str, version: _serializable.Version) -> _serializable.CompositeType:
        if _serializable.CompositeType.NAME_COMPONENT_SEPARATOR in name:
            full_name = name
        else:
            full_name = _serializable.CompositeType.NAME_COMPONENT_SEPARATOR.join([self._definition.full_namespace,
                                                                                   name])
            _logger.debug('The full name of a relatively referred type %r reconstructed as %r', name, full_name)

        del name
        found = list(filter(lambda d: d.full_name == full_name and d.version == version, self._lookup_definitions))
        if not found:
            # Play Sherlock to help the user with mistakes like https://forum.uavcan.org/t/dsdl-compilation-error/904/2
            requested_ns = full_name.split(_serializable.CompositeType.NAME_COMPONENT_SEPARATOR)[0]
            lookup_nss = set(x.root_namespace for x in self._lookup_definitions)
            subroot_ns = self._definition.name_components[1] if len(self._definition.name_components) > 2 else None
            error_description = 'Data type %s.%d.%d could not be found in the following root namespaces: %s. ' % \
                                (full_name, version.major, version.minor, lookup_nss or '(empty set)')
            if requested_ns not in lookup_nss and requested_ns == subroot_ns:
                error_description += ' Did you mean to use the directory %r instead of %r?' % (
                    os.path.join(self._definition.root_namespace_path, subroot_ns),
                    self._definition.root_namespace_path,
                )
            else:
                error_description += ' Please make sure that you specified the directories correctly.'
            raise UndefinedDataTypeError(error_description)

        if len(found) > 1:  # pragma: no cover
            raise _error.InternalError('Conflicting definitions: %r' % found)

        target_definition = found[0]
        assert isinstance(target_definition, _dsdl_definition.DSDLDefinition)
        assert target_definition.full_name == full_name
        assert target_definition.version == version
        # Recursion is cool.
        return target_definition.read(lookup_definitions=self._lookup_definitions,
                                      print_output_handler=self._print_output_handler,
                                      allow_unregulated_fixed_port_id=self._allow_unregulated_fixed_port_id)

    def _on_attribute(self) -> None:
        if self._structs[-1].extent is not None:
            raise InvalidDirectiveError(
                'The extent directive can only be placed after the last attribute definition in the schema. '
                'This is to prevent errors if the extent is dependent on the bit length set of the data schema.'
            )

    def _on_print_directive(self, line_number: int, value: typing.Optional[_expression.Any]) -> None:
        _logger.info('Print directive at %s:%d%s', self._definition.file_path, line_number,
                     (': %s' % value) if value is not None else ' (no value to print)')
        self._print_output_handler(line_number, str(value if value is not None else ''))

    def _on_assert_directive(self, line_number: int, value: typing.Optional[_expression.Any]) -> None:
        if isinstance(value, _expression.Boolean):
            if not value.native_value:
                raise AssertionCheckFailureError('Assertion check has failed',
                                                 path=self._definition.file_path,
                                                 line=line_number)
            else:
                _logger.debug('Assertion check successful at %s:%d', self._definition.file_path, line_number)
        elif value is None:
            raise InvalidDirectiveError('Assert directive requires an expression')
        else:
            raise InvalidDirectiveError('The assertion check expression must yield a boolean, not %s' %
                                        value.TYPE_NAME)

    def _on_extent_directive(self, line_number: int, value: typing.Optional[_expression.Any]) -> None:
        if self._structs[-1].sealed:
            raise InvalidDirectiveError('The extent cannot be specified for a sealed type')
        if value is None:
            raise InvalidDirectiveError('The extent directive requires an expression')
        elif isinstance(value, _expression.Rational):
            struct = self._structs[-1]
            if struct.extent is None:
                struct.set_extent(value.as_native_integer())
                _logger.debug('The extent is set to %d bits at %s:%d',
                              struct.extent, self._definition.file_path, line_number)
            else:
                raise InvalidDirectiveError('Duplicated extent directive (already set to %d bits)' % struct.extent)
        else:
            raise InvalidDirectiveError('The extent directive expects a rational, not %s' % value.TYPE_NAME)

    def _on_sealed_directive(self, _ln: int, value: typing.Optional[_expression.Any]) -> None:
        if self._structs[-1].extent is not None:
            raise InvalidDirectiveError('A type whose extent is defined explicitly cannot be made sealed')
        if value is not None:
            raise InvalidDirectiveError('The sealed directive does not expect an expression')
        if self._structs[-1].sealed:
            raise InvalidDirectiveError('Duplicated sealed directive')
        self._structs[-1].make_sealed()

    def _on_union_directive(self, _ln: int, value: typing.Optional[_expression.Any]) -> None:
        if value is not None:
            raise InvalidDirectiveError('The union directive does not expect an expression')
        if self._structs[-1].union:
            raise InvalidDirectiveError('Duplicated union directive')
        if not self._structs[-1].empty:
            raise InvalidDirectiveError('The union directive must be placed before the first '
                                        'attribute definition')
        self._structs[-1].make_union()

    def _on_deprecated_directive(self, _ln: int, value: typing.Optional[_expression.Any]) -> None:
        if value is not None:
            raise InvalidDirectiveError('The deprecated directive does not expect an expression')
        if self._is_deprecated:
            raise InvalidDirectiveError('Duplicated deprecated directive')
        if len(self._structs) > 1:
            raise InvalidDirectiveError('The deprecated directive cannot be placed in the response section')
        if not self._structs[-1].empty:
            raise InvalidDirectiveError('The deprecated directive must be placed before the first '
                                        'attribute definition')
        self._is_deprecated = True

    @staticmethod
    def _render_explicit_extent_recommendation(model: _serializable.DelimitedType) -> str:
        assert model.extent % 8 == 0, 'Internal error: Extent is expected to be an integer multiple of 8 bits'
        e_bytes = model.extent // 8
        return (
            'Extent is not specified explicitly, defaulting to {e_bytes} bytes. '
            'To accept the default extent and silence this diagnostic, '
            'add the following line near the end of the definition: '
            '`@extent {e_bytes} * 8`'
        ).format(e_bytes=e_bytes)



#
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

from ._any import Any as Any
from ._any import UndefinedOperatorError as UndefinedOperatorError
from ._any import UndefinedAttributeError as UndefinedAttributeError
from ._any import InvalidOperandError as InvalidOperandError

from ._primitive import Primitive as Primitive
from ._primitive import Rational as Rational
from ._primitive import Boolean as Boolean
from ._primitive import String as String

from ._container import Container as Container
from ._container import Set as Set

from ._operator import OperatorOutput as OperatorOutput
from ._operator import BinaryOperator as BinaryOperator
from ._operator import AttributeOperator as AttributeOperator
from ._operator import positive as positive
from ._operator import negative as negative
from ._operator import logical_not as logical_not
from ._operator import logical_or as logical_or
from ._operator import logical_and as logical_and
from ._operator import equal as equal
from ._operator import not_equal as not_equal
from ._operator import less_or_equal as less_or_equal
from ._operator import greater_or_equal as greater_or_equal
from ._operator import less as less
from ._operator import greater as greater
from ._operator import bitwise_and as bitwise_and
from ._operator import bitwise_xor as bitwise_xor
from ._operator import bitwise_or as bitwise_or
from ._operator import add as add
from ._operator import subtract as subtract
from ._operator import multiply as multiply
from ._operator import divide as divide
from ._operator import modulo as modulo
from ._operator import power as power
from ._operator import attribute as attribute



#
# Copyright (C) 2018-2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

from .._bit_length_set import BitLengthSet
from ._serializable import SerializableType
from ._primitive import InvalidBitLengthError


class VoidType(SerializableType):
    MAX_BIT_LENGTH = 64

    def __init__(self, bit_length: int):
        super(VoidType, self).__init__()
        self._bit_length = int(bit_length)

        if self._bit_length < 1:
            raise InvalidBitLengthError('Bit length must be positive')

        if self._bit_length > self.MAX_BIT_LENGTH:
            raise InvalidBitLengthError('Bit length cannot exceed %r' % self.MAX_BIT_LENGTH)

    @property
    def bit_length_set(self) -> BitLengthSet:
        return BitLengthSet(self.bit_length)

    @property
    def bit_length(self) -> int:
        """
        This is a shortcut for ``next(iter(x.bit_length_set))``, because the bit length set of a void type
        always contains exactly one element (i.e., void types are fixed-length).
        """
        return self._bit_length

    @property
    def alignment_requirement(self) -> int:
        return 1

    def __str__(self) -> str:
        return 'void%d' % self.bit_length

    def __repr__(self) -> str:
        return 'VoidType(bit_length=%d)' % self.bit_length


def _unittest_void() -> None:
    from pytest import raises

    assert VoidType(1).bit_length_set == 1
    assert str(VoidType(13)) == 'void13'
    assert repr(VoidType(64)) == 'VoidType(bit_length=64)'
    assert VoidType(22).bit_length_set == {22}
    assert VoidType(1) == VoidType(1)
    assert VoidType(1) != VoidType(2)
    assert hash(VoidType(1)) == hash(VoidType(1))
    assert hash(VoidType(1)) != hash(VoidType(2))
    assert hash(VoidType(1)) != hash(VoidType(11))
    assert hash(VoidType(11)) == hash(VoidType(11))

    with raises(InvalidBitLengthError):
        VoidType(1)
        VoidType(0)

    with raises(InvalidBitLengthError):
        VoidType(64)
        VoidType(65)



#
# Copyright (C) 2018-2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import typing
import functools
from . import _any, _primitive


OperatorOutput    = typing.TypeVar('OperatorOutput')
BinaryOperator    = typing.Callable[[_any.Any, _any.Any], OperatorOutput]
AttributeOperator = typing.Callable[[_any.Any, typing.Union[_primitive.String, str]], OperatorOutput]


#
# Operator wrappers. These wrappers serve two purposes:
#   - Late binding, as explained here: https://stackoverflow.com/questions/55148139/referring-to-a-pure-virtual-method
#   - Automatic left-right operand swapping when necessary (for some polyadic operators).
#
def _auto_swap(alternative_operator_name: typing.Optional[str] = None) -> \
        typing.Callable[[BinaryOperator[OperatorOutput]], BinaryOperator[OperatorOutput]]:
    def decorator(direct_operator: BinaryOperator[OperatorOutput]) -> BinaryOperator[OperatorOutput]:
        if alternative_operator_name:
            alternative_method_name = '_' + alternative_operator_name
        else:
            alternative_method_name = '_%s_right' % direct_operator.__name__

        if not hasattr(_any.Any, alternative_method_name):  # pragma: no cover
            raise TypeError('The following alternative operator method is not defined: %r' % alternative_method_name)

        @functools.wraps(direct_operator)
        def wrapper(left: _any.Any, right: _any.Any) -> _any.Any:
            if not isinstance(left, _any.Any) or not isinstance(right, _any.Any):  # pragma: no cover
                raise ValueError('Operators are only defined for implementations of Any; found this: %r, %r' %
                                 (type(left).__name__, type(right).__name__))
            try:
                result = direct_operator(left, right)
            except _any.UndefinedOperatorError:
                if type(left) != type(right):
                    result = getattr(right, alternative_method_name)(left)  # Left and Right are swapped.
                else:
                    raise

            assert isinstance(result, _any.Any)
            return result
        return wrapper
    return decorator


def logical_not(operand: _any.Any) -> _primitive.Boolean:                       # noinspection PyProtectedMember
    result = operand._logical_not()
    assert isinstance(result, _primitive.Boolean)
    return result


def positive(operand: _any.Any) -> _any.Any:                                    # noinspection PyProtectedMember
    return operand._positive()


def negative(operand: _any.Any) -> _any.Any:                                    # noinspection PyProtectedMember
    return operand._negative()


@_auto_swap('logical_or')  # Commutative
def logical_or(left: _any.Any, right: _any.Any) -> _primitive.Boolean:          # noinspection PyProtectedMember
    result = left._logical_or(right)
    assert isinstance(result, _primitive.Boolean)
    return result


@_auto_swap('logical_and')  # Commutative
def logical_and(left: _any.Any, right: _any.Any) -> _primitive.Boolean:         # noinspection PyProtectedMember
    result = left._logical_and(right)
    assert isinstance(result, _primitive.Boolean)
    return result


@_auto_swap('equal')  # Commutative
def equal(left: _any.Any, right: _any.Any) -> _primitive.Boolean:               # noinspection PyProtectedMember
    result = left._equal(right)
    assert isinstance(result, _primitive.Boolean)
    return result


# Special case - synthetic operator.
def not_equal(left: _any.Any, right: _any.Any) -> _primitive.Boolean:           # noinspection PyProtectedMember
    result = logical_not(equal(left, right))
    assert isinstance(result, _primitive.Boolean)
    return result


@_auto_swap('greater_or_equal')
def less_or_equal(left: _any.Any, right: _any.Any) -> _primitive.Boolean:       # noinspection PyProtectedMember
    result = left._less_or_equal(right)
    assert isinstance(result, _primitive.Boolean)
    return result


@_auto_swap('less_or_equal')
def greater_or_equal(left: _any.Any, right: _any.Any) -> _primitive.Boolean:    # noinspection PyProtectedMember
    result = left._greater_or_equal(right)
    assert isinstance(result, _primitive.Boolean)
    return result


@_auto_swap('greater')
def less(left: _any.Any, right: _any.Any) -> _primitive.Boolean:                # noinspection PyProtectedMember
    result = left._less(right)
    assert isinstance(result, _primitive.Boolean)
    return result


@_auto_swap('less')
def greater(left: _any.Any, right: _any.Any) -> _primitive.Boolean:             # noinspection PyProtectedMember
    result = left._greater(right)
    assert isinstance(result, _primitive.Boolean)
    return result


@_auto_swap()
def bitwise_or(left: _any.Any, right: _any.Any) -> _any.Any:                    # noinspection PyProtectedMember
    return left._bitwise_or(right)


@_auto_swap()
def bitwise_xor(left: _any.Any, right: _any.Any) -> _any.Any:                   # noinspection PyProtectedMember
    return left._bitwise_xor(right)


@_auto_swap()
def bitwise_and(left: _any.Any, right: _any.Any) -> _any.Any:                   # noinspection PyProtectedMember
    return left._bitwise_and(right)


@_auto_swap()
def add(left: _any.Any, right: _any.Any) -> _any.Any:                           # noinspection PyProtectedMember
    return left._add(right)


@_auto_swap()
def subtract(left: _any.Any, right: _any.Any) -> _any.Any:                      # noinspection PyProtectedMember
    return left._subtract(right)


@_auto_swap()
def multiply(left: _any.Any, right: _any.Any) -> _any.Any:                      # noinspection PyProtectedMember
    return left._multiply(right)


@_auto_swap()
def divide(left: _any.Any, right: _any.Any) -> _any.Any:                        # noinspection PyProtectedMember
    return left._divide(right)


@_auto_swap()
def modulo(left: _any.Any, right: _any.Any) -> _any.Any:                        # noinspection PyProtectedMember
    return left._modulo(right)


@_auto_swap()
def power(left: _any.Any, right: _any.Any) -> _any.Any:                         # noinspection PyProtectedMember
    return left._power(right)


# Special case - no argument-swapped alternative defined.
# We accept both native strings and String in order to support both dynamically computed attributes and
# statically defined attributes.
def attribute(value: _any.Any, name: typing.Union[str, _primitive.String]) -> _any.Any:
    if isinstance(name, str):
        name = _primitive.String(name)

    if isinstance(value, _any.Any) and isinstance(name, _primitive.String):     # noinspection PyProtectedMember
        return value._attribute(name)
    else:  # pragma: no cover
        raise ValueError('The argument types of the attribute operator are (Any, String), got (%r, %r)' %
                         (type(value).__name__, type(name).__name__))



#
# Copyright (C) 2018-2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import typing
from . import _error
from . import _serializable
from . import _bit_length_set


class BitLengthAnalysisError(_error.InvalidDefinitionError):
    pass


class DataSchemaBuilder:
    def __init__(self) -> None:
        self._fields = []       # type: typing.List[_serializable.Field]
        self._constants = []    # type: typing.List[_serializable.Constant]
        self._extent = None     # type: typing.Optional[int]
        self._is_sealed = False
        self._is_union = False
        self._bit_length_computed_at_least_once = False

    @property
    def fields(self) -> typing.List[_serializable.Field]:
        assert all(map(lambda x: isinstance(x, _serializable.Field), self._fields))
        return self._fields

    @property
    def constants(self) -> typing.List[_serializable.Constant]:
        assert all(map(lambda x: isinstance(x, _serializable.Constant), self._constants))
        return self._constants

    @property
    def attributes(self) -> typing.List[_serializable.Attribute]:  # noinspection PyTypeChecker
        out = []  # type: typing.List[_serializable.Attribute]
        out += self.fields
        out += self.constants
        return out

    @property
    def empty(self) -> bool:
        return not self._fields and not self._constants

    @property
    def extent(self) -> typing.Optional[int]:
        return self._extent

    @property
    def sealed(self) -> bool:
        return self._is_sealed

    @property
    def union(self) -> bool:
        return self._is_union

    @property
    def offset(self) -> _bit_length_set.BitLengthSet:
        # We set this flag in order to detect invalid reliance on the bit length estimates for unions:
        # we process definitions sequentially, statement-by-statement, so we can't know if there are going to be
        # extra fields added after the bit length values are computed. If we are building a regular structure,
        # this is fine, because in that case each computed value refers to the offset of the next field (if there
        # is one) or the total length of the structure (if we're past the last field). With unions, however,
        # there is no concept of inter-field offset because a union holds exactly one field at any moment;
        # only the total offset (i.e., total size) is defined.
        self._bit_length_computed_at_least_once = True
        ty = _serializable.UnionType if self.union else _serializable.StructureType
        out = ty.aggregate_bit_length_sets([f.data_type for f in self.fields])  # type: ignore
        assert isinstance(out, _bit_length_set.BitLengthSet) and len(out) > 0
        return out

    def add_field(self, field: _serializable.Field) -> None:
        if self.union and self._bit_length_computed_at_least_once:
            # Refer to the DSDL specification for the background information.
            raise BitLengthAnalysisError('Inter-field offset is not defined for unions; '
                                         'previously performed bit length analysis is invalid')
        assert isinstance(field, _serializable.Field)
        self._fields.append(field)

    def add_constant(self, constant: _serializable.Constant) -> None:
        assert isinstance(constant, _serializable.Constant)
        self._constants.append(constant)

    def set_extent(self, value: int) -> None:
        assert self._extent is None
        assert not self._is_sealed
        self._extent = int(value)

    def make_sealed(self) -> None:
        assert not self.sealed
        assert self._extent is None
        self._is_sealed = True

    def make_union(self) -> None:
        assert not self.union
        self._is_union = True

    def to_service_schema_params(self) -> _serializable.ServiceType.SchemaParams:
        return _serializable.ServiceType.SchemaParams(
            attributes=self.attributes,
            extent=self.extent,
            is_sealed=self.sealed,
            is_union=self.union,
        )



#
# Copyright (C) 2018-2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import typing
import itertools


class BitLengthSet:
    """
    This type represents the Bit Length Set as defined in the Specification.
    It is used for representing bit offsets of fields and bit lengths of serialized representations.

    Instances are comparable between each other, with plain integers, and with native sets of integers.
    The methods do not mutate the instance they are invoked on; instead, the result is returned as a new instance,
    excepting the in-place ``__ixx__()`` operator overloads.
    """

    def __init__(self, values: typing.Optional[typing.Union[typing.Iterable[int], int]] = None):
        """
        Accepts any iterable that yields integers (like another bit length set) or a single integer,
        in which case it will result in the set containing only the one specified integer.
        The source container is always deep-copied.

        >>> BitLengthSet()
        BitLengthSet()
        >>> len(BitLengthSet()) == 0
        True
        >>> BitLengthSet(1)
        BitLengthSet({1})
        >>> BitLengthSet({1, 2, 3})
        BitLengthSet({1, 2, 3})
        """
        if values is None:
            values = set()
        elif isinstance(values, int):
            values = {values}
        else:
            values = set(map(int, values))

        if not all(map(lambda x: x >= 0, values)):
            raise ValueError('Bit length set elements cannot be negative: %r' % values)

        assert isinstance(values, set)
        assert all(map(lambda x: isinstance(x, int) and x >= 0, values))
        self._value = values  # type: typing.Set[int]

    def is_aligned_at(self, bit_length: int) -> bool:
        """
        Checks whether all of the contained offset values match the specified alignment goal.
        An empty bit length set is considered to have infinite alignment.

        >>> BitLengthSet(64).is_aligned_at(32)
        True
        >>> BitLengthSet(48).is_aligned_at(32)
        False
        >>> BitLengthSet(48).is_aligned_at(16)
        True
        >>> BitLengthSet().is_aligned_at(123456)
        True
        """
        if self:
            return set(map(lambda x: x % bit_length, self._value)) == {0}
        else:
            return True     # An empty set is always aligned.

    def is_aligned_at_byte(self) -> bool:
        """
        A shorthand for :meth:`is_aligned_at` using the standard byte size as prescribed by the Specification.

        >>> BitLengthSet(32).is_aligned_at_byte()
        True
        >>> BitLengthSet(33).is_aligned_at_byte()
        False
        """
        from ._serializable import SerializableType
        return self.is_aligned_at(SerializableType.BITS_PER_BYTE)

    def pad_to_alignment(self, bit_length: int) -> 'BitLengthSet':
        """
        Pad each element in the set such that the set becomes aligned at the specified alignment goal.
        After this transformation is applied, elements may become up to ``bit_length-1`` bits larger.
        The argument shall be a positive integer, otherwise it's a :class:`ValueError`.

        >>> BitLengthSet({0, 1, 2, 3, 4, 5, 6, 7, 8}).pad_to_alignment(1)  # Alignment to 1 is a no-op.
        BitLengthSet({0, 1, 2, 3, 4, 5, 6, 7, 8})
        >>> BitLengthSet({0, 1, 2, 3, 4, 5, 6, 7, 8}).pad_to_alignment(2)
        BitLengthSet({0, 2, 4, 6, 8})
        >>> BitLengthSet({0, 1, 5, 7}).pad_to_alignment(2)
        BitLengthSet({0, 2, 6, 8})
        >>> BitLengthSet({0, 1, 2, 3, 4, 5, 6, 7, 8}).pad_to_alignment(3)
        BitLengthSet({0, 3, 6, 9})
        >>> BitLengthSet({0, 1, 2, 3, 4, 5, 6, 7, 8}).pad_to_alignment(8)
        BitLengthSet({0, 8})
        >>> BitLengthSet({0, 9}).pad_to_alignment(8)
        BitLengthSet({0, 16})
        >>> from random import randint
        >>> alignment = randint(1, 64)
        >>> BitLengthSet(randint(1, 1000) for _ in range(100)).pad_to_alignment(alignment).is_aligned_at(alignment)
        True
        """
        r = int(bit_length)
        if r < 1:
            raise ValueError('Invalid alignment: %r bits' % r)
        else:
            assert r >= 1
            out = BitLengthSet(((x + r - 1) // r) * r for x in self)
            assert not out or 0 <= min(out) - min(self) < r
            assert not out or 0 <= max(out) - max(self) < r
            assert len(out) <= len(self)
            return out

    def elementwise_sum_k_multicombinations(self, k: int) -> 'BitLengthSet':
        """
        This is a special case of :meth:`elementwise_sum_cartesian_product`.
        The original object is not modified.

        One can replace this method with the aforementioned general case method and the behavior would not change;
        however, we need this special case method for performance reasons. When dealing with arrays (either fixed- or
        variable-length), usage of this method instead of the generic one yields significantly better performance,
        since the computational complexity of k-selections is much lower than that of the Cartesian product.

        >>> BitLengthSet(1).elementwise_sum_k_multicombinations(1)
        BitLengthSet({1})
        >>> BitLengthSet({1, 2, 3}).elementwise_sum_k_multicombinations(1)
        BitLengthSet({1, 2, 3})
        >>> BitLengthSet({1, 2, 3}).elementwise_sum_k_multicombinations(2)
        BitLengthSet({2, 3, 4, 5, 6})
        """
        k_multicombination = itertools.combinations_with_replacement(self, k)
        elementwise_sums = map(sum, k_multicombination)
        return BitLengthSet(elementwise_sums)  # type: ignore

    @staticmethod
    def elementwise_sum_cartesian_product(sets: typing.Iterable[typing.Union[typing.Iterable[int], int]]) \
            -> 'BitLengthSet':
        """
        This operation is fundamental for bit length and bit offset (which are, generally, the same thing) computation.

        The basic background is explained in the specification. The idea is that the bit offset of a given entity
        in a data type definition of the structure category (or, in other words, the bit length set of serialized
        representations of the preceding entities, which is the same thing, assuming that the data type is of the
        structure category) is a function of bit length sets of each preceding entity. Various combinations of
        bit lengths of the preceding entities are possible, which can be expressed through the Cartesian product over
        the bit length sets of the preceding entities. Since in a type of the structure category entities are arranged
        as an ordered sequence of a fixed length (meaning that entities can't be added or removed), the resulting
        bit length (offset) is computed by elementwise summation of each element of the Cartesian product.

        This method is not applicable for the tagged union type category, since a tagged union holds exactly one
        value at any moment; therefore, the bit length set of a tagged union is simply a union of bit length sets
        of each entity that can be contained in the union, plus the length of the implicit union tag field.

        From the standpoint of bit length combination analysis, fixed-length arrays are a special case of structures,
        because they also contain a fixed ordered sequence of fields, where all fields are of the same type.
        The method defined for structures applies to fixed-length arrays, but one should be aware that it may be
        computationally suboptimal, since the fact that all array elements are of the same type allows us to replace
        the computationally expensive Cartesian product with k-multicombinations (k-selections).

        In the context of bit length analysis, variable-length arrays do not require any special treatment, since a
        variable-length array with the capacity of N elements can be modeled as a tagged union containing
        N fixed arrays of length from 1 to N, plus one empty field (representing the case of an empty variable-length
        array).

        >>> BitLengthSet.elementwise_sum_cartesian_product([1, 2, 10])
        BitLengthSet({13})
        >>> BitLengthSet.elementwise_sum_cartesian_product([{1, 2}, {4, 5}])
        BitLengthSet({5, 6, 7})
        >>> BitLengthSet.elementwise_sum_cartesian_product([{1, 2, 3}, {4, 5, 6}])
        BitLengthSet({5, 6, 7, 8, 9})
        >>> BitLengthSet.elementwise_sum_cartesian_product([{1, 2, 3}, {4, 5, 6}, {7, 8, 9}])
        BitLengthSet({12, 13, 14, 15, 16, 17, 18})
        """
        cartesian_product = itertools.product(*list(map(BitLengthSet, sets)))
        elementwise_sums = map(sum, cartesian_product)
        return BitLengthSet(elementwise_sums)  # type: ignore

    def __iter__(self) -> typing.Iterator[int]:
        return iter(self._value)

    def __len__(self) -> int:
        """Cardinality."""
        return len(self._value)

    def __eq__(self, other: typing.Any) -> bool:
        """
        Whether the current set equals the other.
        The other may be a bit length set, an integer, or a native ``typing.Set[int]``.
        """
        if isinstance(other, _OPERAND_TYPES):
            return self._value == BitLengthSet(other)._value
        else:
            return NotImplemented

    def __bool__(self) -> bool:
        """
        Evaluates to True unless empty.

        >>> assert not BitLengthSet()
        >>> assert not BitLengthSet({})
        >>> assert BitLengthSet(0)
        >>> assert BitLengthSet({1, 2, 3})
        """
        return bool(self._value)

    def __add__(self, other: typing.Any) -> 'BitLengthSet':
        """
        This operation models the addition of a new object to a serialized representation;
        i.e., it is an alias for ``elementwise_sum_cartesian_product([self, other])``.
        The result is stored into a new instance which is returned.

        If the argument is a bit length set, an elementwise sum set of the Cartesian product of the argument set
        with the current set will be computed, and the result will be returned as a new set (self is not modified).
        One can easily see that if the argument is a set of one value (or a scalar),
        this method will result in the addition of said scalar to every element of the original set.

        SPECIAL CASE: if the current set is empty at the time of invocation, it will be assumed to be equal ``{0}``.

        The other may be a bit length set, an integer, or a native ``typing.Set[int]``.

        >>> BitLengthSet() + BitLengthSet()
        BitLengthSet()
        >>> BitLengthSet(4) + BitLengthSet(3)
        BitLengthSet({7})
        >>> BitLengthSet({4, 91}) + 3
        BitLengthSet({7, 94})
        >>> BitLengthSet({4, 91}) + {5, 7}
        BitLengthSet({9, 11, 96, 98})
        """
        if isinstance(other, _OPERAND_TYPES):
            return BitLengthSet.elementwise_sum_cartesian_product([self or BitLengthSet(0), BitLengthSet(other)])
        else:
            return NotImplemented

    def __radd__(self, other: typing.Any) -> 'BitLengthSet':
        """
        See :meth:`__add__`.

        >>> {1, 2, 3} + BitLengthSet({4, 5, 6})
        BitLengthSet({5, 6, 7, 8, 9})
        >>> 1 + BitLengthSet({2, 5, 7})
        BitLengthSet({3, 6, 8})
        """
        if isinstance(other, _OPERAND_TYPES):
            return BitLengthSet(other) + self
        else:
            return NotImplemented

    def __iadd__(self, other: typing.Any) -> 'BitLengthSet':
        """
        See :meth:`__add__`.

        >>> a = BitLengthSet({1, 2, 3})
        >>> a += {4, 5, 6}
        >>> a
        BitLengthSet({5, 6, 7, 8, 9})
        """
        if isinstance(other, _OPERAND_TYPES):
            self._value = (self + other)._value
            return self
        else:
            return NotImplemented

    def __or__(self, other: typing.Any) -> 'BitLengthSet':
        """
        Creates and returns a new set that is a union of this set with another bit length set.

        >>> a = BitLengthSet()
        >>> a = a | BitLengthSet({1, 2, 3})
        >>> a
        BitLengthSet({1, 2, 3})
        >>> a = a | {3, 4, 5}
        >>> a
        BitLengthSet({1, 2, 3, 4, 5})
        >>> a | 6
        BitLengthSet({1, 2, 3, 4, 5, 6})
        """
        if isinstance(other, _OPERAND_TYPES):
            return BitLengthSet(self._value | BitLengthSet(other)._value)
        else:
            return NotImplemented

    def __ror__(self, other: typing.Any) -> 'BitLengthSet':
        """
        See :meth:`__or__`.

        >>> {1, 2, 3} | BitLengthSet({4, 5, 6})
        BitLengthSet({1, 2, 3, 4, 5, 6})
        >>> 1 | BitLengthSet({2, 5, 7})
        BitLengthSet({1, 2, 5, 7})
        """
        if isinstance(other, _OPERAND_TYPES):
            return BitLengthSet(other) | self
        else:
            return NotImplemented

    def __ior__(self, other: typing.Any) -> 'BitLengthSet':
        """
        See :meth:`__or__`.

        >>> a = BitLengthSet({4, 5, 6})
        >>> a |= {1, 2, 3}
        >>> a
        BitLengthSet({1, 2, 3, 4, 5, 6})
        """
        if isinstance(other, _OPERAND_TYPES):
            self._value = (self | other)._value
            return self
        else:
            return NotImplemented

    def __str__(self) -> str:
        """
        Always yields a sorted representation for the ease of human consumption.

        >>> str(BitLengthSet())
        '{}'
        >>> str(BitLengthSet({918, 16, 7, 42}))
        '{7, 16, 42, 918}'
        """
        return '{' + ', '.join(map(str, sorted(self._value))) + '}'

    def __repr__(self) -> str:
        """
        >>> BitLengthSet()
        BitLengthSet()
        >>> BitLengthSet({918, 16, 7, 42})
        BitLengthSet({7, 16, 42, 918})
        """
        return type(self).__name__ + '(' + str(self or '') + ')'


_OPERAND_TYPES = BitLengthSet, set, int


def _unittest_bit_length_set() -> None:
    from pytest import raises
    assert not BitLengthSet()
    assert BitLengthSet() == BitLengthSet()
    assert not (BitLengthSet() != BitLengthSet())
    assert BitLengthSet(123) == BitLengthSet([123])
    assert BitLengthSet(123) != BitLengthSet(124)
    assert BitLengthSet(123) == 123
    assert BitLengthSet(123) != 124
    assert not (BitLengthSet(123) == '123')  # not implemented
    assert str(BitLengthSet()) == '{}'
    assert str(BitLengthSet(123)) == '{123}'
    assert str(BitLengthSet((123, 0, 456, 12))) == '{0, 12, 123, 456}'  # Always sorted!
    assert BitLengthSet().is_aligned_at(1)
    assert BitLengthSet().is_aligned_at(1024)
    assert BitLengthSet(8).is_aligned_at_byte()
    assert not BitLengthSet(8).is_aligned_at(16)
    with raises(ValueError):
        BitLengthSet({-1})

    s = BitLengthSet({0, 8})
    s += 8
    assert s == {8, 16}
    s = s + {0, 4, 8}
    assert s == {8, 16, 12, 20, 24}

    assert BitLengthSet() + BitLengthSet() == BitLengthSet()
    assert BitLengthSet(4) + BitLengthSet(3) == {7}
    assert BitLengthSet({4, 91}) + 3 == {7, 94}
    assert BitLengthSet(7) + {12, 15} == {19, 22}
    assert {1, 2, 3} + BitLengthSet([4, 5, 6]) == {5, 6, 7, 8, 9}

    with raises(TypeError):
        assert BitLengthSet([4, 5, 6]) + '1'

    with raises(TypeError):
        assert '1' + BitLengthSet([4, 5, 6])

    with raises(TypeError):
        s = BitLengthSet([4, 5, 6])
        s += '1'

    with raises(TypeError):
        assert '1' | BitLengthSet([4, 5, 6])

    with raises(TypeError):
        assert BitLengthSet([4, 5, 6]) | '1'

    with raises(TypeError):
        s = BitLengthSet([4, 5, 6])
        s |= '1'

    with raises(ValueError):
        BitLengthSet([4, 5, 6]).pad_to_alignment(0)



#
# Copyright (C) 2018-2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

MAX_SUBJECT_ID = 8191
MAX_SERVICE_ID = 511


_STANDARD_ROOT_NAMESPACE = 'uavcan'

_STANDARD_MESSAGES = 7168, 8191
_STANDARD_SERVICES = 384, 511

# The upper end of the ranges may be consumed by the standard types shall that become necessary,
# so new fixed port-ID allocations should be granted at the bottom of the ranges only.
_VENDOR_MESSAGES = 6144, 7167
_VENDOR_SERVICES = 256, 383


def is_valid_regulated_subject_id(regulated_id: int,
                                  root_namespace: str) -> bool:
    is_standard = root_namespace.strip() == _STANDARD_ROOT_NAMESPACE
    lo, hi = _STANDARD_MESSAGES if is_standard else _VENDOR_MESSAGES
    return lo <= int(regulated_id) <= hi


def is_valid_regulated_service_id(regulated_id: int,
                                  root_namespace: str) -> bool:
    is_standard = root_namespace.strip() == _STANDARD_ROOT_NAMESPACE
    lo, hi = _STANDARD_SERVICES if is_standard else _VENDOR_SERVICES
    return lo <= int(regulated_id) <= hi


def _unittest_pid_ranges() -> None:
    assert is_valid_regulated_subject_id(regulated_id=7000, root_namespace='sirius_cybernetics_corp')
    assert not is_valid_regulated_subject_id(regulated_id=7000, root_namespace='uavcan')
    assert is_valid_regulated_subject_id(regulated_id=8000, root_namespace='uavcan')
    assert not is_valid_regulated_subject_id(regulated_id=8000, root_namespace='sirius_cybernetics_corp')
    assert not is_valid_regulated_subject_id(regulated_id=6000, root_namespace='uavcan')
    assert not is_valid_regulated_subject_id(regulated_id=6000, root_namespace='sirius_cybernetics_corp')

    assert is_valid_regulated_service_id(regulated_id=260, root_namespace='sirius_cybernetics_corp')
    assert not is_valid_regulated_service_id(regulated_id=260, root_namespace='uavcan')
    assert is_valid_regulated_service_id(regulated_id=400, root_namespace='uavcan')
    assert not is_valid_regulated_service_id(regulated_id=400, root_namespace='sirius_cybernetics_corp')
    assert not is_valid_regulated_service_id(regulated_id=600, root_namespace='uavcan')
    assert not is_valid_regulated_service_id(regulated_id=600, root_namespace='sirius_cybernetics_corp')



#
# Copyright (C) 2018-2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import abc
import typing
import functools
from . import _any, _primitive, _operator


_O = typing.TypeVar('_O')


# noinspection PyAbstractClass
class Container(_any.Any):
    @property
    @abc.abstractmethod
    def element_type(self) -> typing.Type[_any.Any]:
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def __iter__(self) -> typing.Iterator[typing.Any]:
        raise NotImplementedError  # pragma: no cover


class Set(Container):
    TYPE_NAME = 'set'

    # noinspection PyProtectedMember
    class _Decorator:
        @staticmethod
        def homotypic_binary_operator(inferior: typing.Callable[['Set', 'Set'], _O]) \
                -> typing.Callable[['Set', 'Set'], _O]:
            def wrapper(self: 'Set', other: 'Set') -> _O:
                assert isinstance(self, Set) and isinstance(other, Set)
                if self.element_type == other.element_type:
                    return inferior(self, other)
                else:
                    raise _any.InvalidOperandError('The requested binary operator is defined only for sets '
                                                   'that share the same element type. The different types are: %r, %r' %
                                                   (self.element_type.TYPE_NAME, other.element_type.TYPE_NAME))
            return wrapper

    def __init__(self, elements: typing.Iterable[_any.Any]):
        list_of_elements = list(elements)   # type: typing.List[_any.Any]
        del elements
        if len(list_of_elements) < 1:
            raise _any.InvalidOperandError('Zero-length sets are currently not permitted because '
                                           'of associated type deduction issues. This may change later.')

        element_types = set(map(type, list_of_elements))
        if len(element_types) != 1:
            # This also weeds out covariant sets, although our barbie-size type system is unaware of that.
            raise _any.InvalidOperandError('Heterogeneous sets are not permitted')

        # noinspection PyTypeChecker
        self._element_type = list(element_types)[0]  # type: typing.Type[_any.Any]
        self._value = frozenset(list_of_elements)    # type: typing.FrozenSet[_any.Any]

        if not issubclass(self._element_type, _any.Any):
            raise ValueError('Invalid element type: %r' % self._element_type)

    def __iter__(self) -> typing.Iterator[typing.Any]:
        return iter(self._value)

    @property
    def element_type(self) -> typing.Type[_any.Any]:
        return self._element_type

    def __hash__(self) -> int:
        return hash(self._value)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Set):
            return self._value == other._value
        else:
            return NotImplemented

    def __str__(self) -> str:
        return '{%s}' % ', '.join(map(str, self._value))    # This is recursive.

    #
    # Set algebra implementation.
    #
    @_Decorator.homotypic_binary_operator
    def _is_equal_to(self, right: 'Set') -> bool:
        return self._value == right._value

    @_Decorator.homotypic_binary_operator
    def _is_superset_of(self, right: 'Set') -> bool:
        return self._value.issuperset(right._value)

    @_Decorator.homotypic_binary_operator
    def _is_subset_of(self, right: 'Set') -> bool:
        return self._value.issubset(right._value)

    @_Decorator.homotypic_binary_operator
    def _is_proper_superset_of(self, right: 'Set') -> bool:
        return self._is_superset_of(right) and not self._is_equal_to(right)

    @_Decorator.homotypic_binary_operator
    def _is_proper_subset_of(self, right: 'Set') -> bool:
        return self._is_subset_of(right) and not self._is_equal_to(right)

    @_Decorator.homotypic_binary_operator
    def _create_union_with(self, right: 'Set') -> 'Set':
        return Set(self._value.union(right._value))

    @_Decorator.homotypic_binary_operator
    def _create_intersection_with(self, right: 'Set') -> 'Set':
        return Set(self._value.intersection(right._value))

    @_Decorator.homotypic_binary_operator
    def _create_disjunctive_union_with(self, right: 'Set') -> 'Set':
        return Set(self._value.symmetric_difference(right._value))

    #
    # Set comparison.
    #
    def _equal(self, right: _any.Any) -> _primitive.Boolean:
        if isinstance(right, Set):
            return _primitive.Boolean(self._is_equal_to(right))
        else:
            raise _any.UndefinedOperatorError

    def _less_or_equal(self, right: _any.Any) -> _primitive.Boolean:
        if isinstance(right, Set):
            return _primitive.Boolean(self._is_subset_of(right))
        else:
            raise _any.UndefinedOperatorError

    def _greater_or_equal(self, right: _any.Any) -> _primitive.Boolean:
        if isinstance(right, Set):
            return _primitive.Boolean(self._is_superset_of(right))
        else:
            raise _any.UndefinedOperatorError

    def _less(self, right: _any.Any) -> _primitive.Boolean:
        if isinstance(right, Set):
            return _primitive.Boolean(self._is_proper_subset_of(right))
        else:
            raise _any.UndefinedOperatorError

    def _greater(self, right: _any.Any) -> _primitive.Boolean:
        if isinstance(right, Set):
            return _primitive.Boolean(self._is_proper_superset_of(right))
        else:
            raise _any.UndefinedOperatorError

    #
    # Set algebra operators that yield a new set.
    #
    def _bitwise_or(self, right: _any.Any) -> 'Set':
        if isinstance(right, Set):
            return self._create_union_with(right)
        else:
            raise _any.UndefinedOperatorError

    def _bitwise_xor(self, right: _any.Any) -> 'Set':
        if isinstance(right, Set):
            return self._create_disjunctive_union_with(right)
        else:
            raise _any.UndefinedOperatorError

    def _bitwise_and(self, right: _any.Any) -> 'Set':
        if isinstance(right, Set):
            return self._create_intersection_with(right)
        else:
            raise _any.UndefinedOperatorError

    #
    # Elementwise application.
    # https://stackoverflow.com/questions/55148139/referring-to-a-pure-virtual-method
    #
    def _elementwise(self,
                     impl: typing.Callable[[_any.Any, _any.Any], _any.Any],
                     other: _any.Any,
                     swap: bool = False) -> 'Set':
        if not isinstance(other, Set):
            return Set((impl(other, x) if swap else impl(x, other)) for x in self)
        else:
            raise _any.UndefinedOperatorError

    def _add(self, right: _any.Any) -> 'Set':
        return self._elementwise(_operator.add, right)

    def _add_right(self, left: _any.Any) -> 'Set':
        return self._elementwise(_operator.add, left, swap=True)

    def _subtract(self, right: _any.Any) -> 'Set':
        return self._elementwise(_operator.subtract, right)

    def _subtract_right(self, left: _any.Any) -> 'Set':
        return self._elementwise(_operator.subtract, left, swap=True)

    def _multiply(self, right: _any.Any) -> 'Set':
        return self._elementwise(_operator.multiply, right)

    def _multiply_right(self, left: _any.Any) -> 'Set':
        return self._elementwise(_operator.multiply, left, swap=True)

    def _divide(self, right: _any.Any) -> 'Set':
        return self._elementwise(_operator.divide, right)

    def _divide_right(self, left: _any.Any) -> 'Set':
        return self._elementwise(_operator.divide, left, swap=True)

    def _modulo(self, right: _any.Any) -> 'Set':
        return self._elementwise(_operator.modulo, right)

    def _modulo_right(self, left: _any.Any) -> 'Set':
        return self._elementwise(_operator.modulo, left, swap=True)

    def _power(self, right: _any.Any) -> 'Set':
        return self._elementwise(_operator.power, right)

    def _power_right(self, left: _any.Any) -> 'Set':
        return self._elementwise(_operator.power, left, swap=True)

    #
    # Attributes
    #
    def _attribute(self, name: '_primitive.String') -> _any.Any:
        if name.native_value == 'min':
            out = functools.reduce(lambda a, b: a if _operator.less(a, b) else b, self)
            assert isinstance(out, self.element_type)
        elif name.native_value == 'max':
            out = functools.reduce(lambda a, b: a if _operator.greater(a, b) else b, self)
            assert isinstance(out, self.element_type)
        elif name.native_value == 'count':  # "size" and "length" can be ambiguous, "cardinality" is long
            out = _primitive.Rational(len(self._value))
        else:
            out = super(Set, self)._attribute(name)  # Hand over up the inheritance chain, this is important

        assert isinstance(out, _any.Any)
        return out



#
# Copyright (C) 2018-2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import abc
import math
import typing
import itertools
import fractions
from .. import _expression
from .. import _port_id_ranges
from .._bit_length_set import BitLengthSet
from ._serializable import SerializableType, TypeParameterError
from ._attribute import Attribute, Field, PaddingField, Constant
from ._name import check_name, InvalidNameError
from ._void import VoidType
from ._primitive import PrimitiveType, UnsignedIntegerType


Version = typing.NamedTuple('Version', [('major', int), ('minor', int)])


class InvalidVersionError(TypeParameterError):
    pass


class AttributeNameCollisionError(TypeParameterError):
    pass


class InvalidExtentError(TypeParameterError):
    pass


class InvalidFixedPortIDError(TypeParameterError):
    pass


class MalformedUnionError(TypeParameterError):
    pass


class DeprecatedDependencyError(TypeParameterError):
    pass


class CompositeType(SerializableType):
    """
    This is the most interesting type in the library because it represents an actual DSDL definition upon its
    interpretation.
    This is an abstract class with several specializations.
    """

    MAX_NAME_LENGTH = 50
    MAX_VERSION_NUMBER = 255
    NAME_COMPONENT_SEPARATOR = '.'

    def __init__(self,
                 name:             str,
                 version:          Version,
                 attributes:       typing.Iterable[Attribute],
                 deprecated:       bool,
                 fixed_port_id:    typing.Optional[int],
                 source_file_path: str,
                 parent_service:   typing.Optional['ServiceType'] = None):
        super(CompositeType, self).__init__()

        self._name = str(name).strip()
        self._version = version
        self._attributes = list(attributes)
        self._attributes_by_name = {a.name: a for a in self._attributes if not isinstance(a, PaddingField)}
        self._deprecated = bool(deprecated)
        self._fixed_port_id = None if fixed_port_id is None else int(fixed_port_id)
        self._source_file_path = str(source_file_path)
        self._parent_service = parent_service

        if self._parent_service is not None:
            assert self._name.endswith('.Request') or self._name.endswith('.Response')
            if not isinstance(self._parent_service, ServiceType):  # pragma: no cover
                raise ValueError('The parent service reference is invalid: %s' % type(parent_service).__name__)

        # Name check
        if not self._name:
            raise InvalidNameError('Composite type name cannot be empty')

        if self.NAME_COMPONENT_SEPARATOR not in self._name:
            raise InvalidNameError('Root namespace is not specified')

        # Do not check name length for synthesized types.
        if len(self._name) > self.MAX_NAME_LENGTH and parent_service is None:
            raise InvalidNameError('Name is too long: %r is longer than %d characters' %
                                   (self._name, self.MAX_NAME_LENGTH))

        for component in self._name.split(self.NAME_COMPONENT_SEPARATOR):
            check_name(component)

        # Version check
        version_valid = (0 <= self._version.major <= self.MAX_VERSION_NUMBER) and\
                        (0 <= self._version.minor <= self.MAX_VERSION_NUMBER) and\
                        ((self._version.major + self._version.minor) > 0)

        if not version_valid:
            raise InvalidVersionError('Invalid version numbers: %s.%s' % (self._version.major, self._version.minor))

        # Attribute check
        used_names = set()      # type: typing.Set[str]
        for a in self._attributes:
            if a.name and a.name in used_names:
                raise AttributeNameCollisionError('Multiple attributes under the same name: %r' % a.name)
            else:
                used_names.add(a.name)

        # Port ID check
        port_id = self._fixed_port_id
        if port_id is not None:
            assert port_id is not None
            if isinstance(self, ServiceType):
                if not (0 <= port_id <= _port_id_ranges.MAX_SERVICE_ID):
                    raise InvalidFixedPortIDError('Fixed service ID %r is not valid' % port_id)
            else:
                if not (0 <= port_id <= _port_id_ranges.MAX_SUBJECT_ID):
                    raise InvalidFixedPortIDError('Fixed subject ID %r is not valid' % port_id)

        # Consistent deprecation check.
        # A non-deprecated type cannot be dependent on deprecated types.
        # A deprecated type can be dependent on anything.
        if not self.deprecated:
            for a in self._attributes:
                t = a.data_type
                if isinstance(t, CompositeType):
                    if t.deprecated:
                        raise DeprecatedDependencyError('A type cannot depend on deprecated types '
                                                        'unless it is also deprecated.')

    @property
    def full_name(self) -> str:
        """The full name, e.g., ``uavcan.node.Heartbeat``."""
        return self._name

    @property
    def name_components(self) -> typing.List[str]:
        """Components of the full name as a list, e.g., ``['uavcan', 'node', 'Heartbeat']``."""
        return self._name.split(CompositeType.NAME_COMPONENT_SEPARATOR)

    @property
    def short_name(self) -> str:
        """The last component of the full name, e.g., ``Heartbeat`` of ``uavcan.node.Heartbeat``."""
        return self.name_components[-1]

    @property
    def full_namespace(self) -> str:
        """The full name without the short name, e.g., ``uavcan.node`` for ``uavcan.node.Heartbeat``."""
        return str(CompositeType.NAME_COMPONENT_SEPARATOR.join(self.name_components[:-1]))

    @property
    def root_namespace(self) -> str:
        """The first component of the full name, e.g., ``uavcan`` of ``uavcan.node.Heartbeat``."""
        return self.name_components[0]

    @property
    def version(self) -> Version:
        """The version numbers of the type, e.g., ``(1, 0)`` of ``uavcan.node.Heartbeat.1.0``."""
        return self._version

    @property
    def extent(self) -> int:
        """
        The amount of memory, in bits, that needs to be allocated in order to store a serialized representation of
        this type or any of its minor versions under the same major version.
        This value is always at least as large as the sum of maximum bit lengths of all fields padded to one byte.
        If the type is sealed, its extent equals ``max(bit_length_set)``.
        """
        return max(self.bit_length_set or {0})

    @property
    def bit_length_set(self) -> BitLengthSet:
        """
        The bit length set of a composite is always aligned at :attr:`alignment_requirement`.
        For a sealed type this is the true bit length set computed by aggregating the fields and
        padding the result to :attr:`alignment_requirement`.
        That is, sealed types expose their internal structure; for example, a type that contains a single field
        of type ``uint32[2]`` would have a single entry in the bit length set: ``{64}``.
        """
        raise NotImplementedError

    @property
    def deprecated(self) -> bool:
        """Whether the definition is marked ``@deprecated``."""
        return self._deprecated

    @property
    def attributes(self) -> typing.List[Attribute]:
        return self._attributes[:]  # Return copy to prevent mutation

    @property
    def fields(self) -> typing.List[Field]:
        return [a for a in self.attributes if isinstance(a, Field)]

    @property
    def fields_except_padding(self) -> typing.List[Field]:
        return [a for a in self.attributes if isinstance(a, Field) and not isinstance(a, PaddingField)]

    @property
    def constants(self) -> typing.List[Constant]:
        return [a for a in self.attributes if isinstance(a, Constant)]

    @property
    def fixed_port_id(self) -> typing.Optional[int]:
        return self._fixed_port_id

    @property
    def has_fixed_port_id(self) -> bool:
        return self.fixed_port_id is not None

    @property
    def source_file_path(self) -> str:
        """
        For synthesized types such as service request/response sections, this property is defined as an empty string.
        """
        return self._source_file_path

    @property
    def alignment_requirement(self) -> int:
        # This is more general than required by the Specification, but it is done this way in case if we decided
        # to support greater alignment requirements in the future.
        return max([self.BITS_PER_BYTE] + [x.data_type.alignment_requirement for x in self.fields])

    @property
    def parent_service(self) -> typing.Optional['ServiceType']:
        """
        :class:`pydsdl.ServiceType` contains two special fields of this type: ``request`` and ``response``.
        For them this property points to the parent service instance; otherwise it's None.
        """
        return self._parent_service

    @abc.abstractmethod
    def iterate_fields_with_offsets(self, base_offset: typing.Optional[BitLengthSet] = None) \
            -> typing.Iterator[typing.Tuple[Field, BitLengthSet]]:
        """
        Iterates over every field (not attribute -- constants are excluded) of the data type,
        yielding it together with its offset, where the offset is represented as :class:`pydsdl.BitLengthSet`.
        The offset of each field is added to the base offset, which may be specified by the caller;
        if not specified, the base offset is assumed to be ``{0}``.

        The objective of this method is to allow code generators to easily implement fully unrolled serialization and
        deserialization routines, where "unrolled" means that upon encountering another (nested) composite type, the
        serialization routine would not delegate its serialization to the serialization routine of the encountered type,
        but instead would serialize it in-place, as if the field of that type was replaced with its own fields in-place.
        The lack of delegation has very important performance implications: when the serialization routine does
        not delegate serialization of the nested types, it can perform infinitely deep field alignment analysis,
        thus being able to reliably statically determine whether each field of the type, including nested types
        at arbitrarily deep levels of nesting, is aligned relative to the origin of the serialized representation
        of the outermost type. As a result, the code generator will be able to avoid unnecessary reliance on slow
        bit-level copy routines replacing them instead with much faster byte-level copy (like ``memcpy()``) or even
        plain memory aliasing.

        When invoked on a tagged union type, the method yields the same offset for every field (since that's how
        tagged unions are serialized), where the offset equals the bit length of the implicit union tag (plus the
        base offset, of course, if provided).

        Please refer to the usage examples to see how this feature can be used.

        :param base_offset: Assume the specified base offset; assume zero offset if the parameter is not provided.
            The base offset will be implicitly padded out to :attr:`alignment_requirement`.

        :return: A generator of ``(Field, BitLengthSet)``.
            Each instance of :class:`pydsdl.BitLengthSet` yielded by the generator is a dedicated copy,
            meaning that the consumer can mutate the returned instances arbitrarily without affecting future values.
            It is guaranteed that each yielded instance is non-empty.
        """
        raise NotImplementedError

    def _attribute(self, name: _expression.String) -> _expression.Any:
        """This is the handler for DSDL expressions like ``uavcan.node.Heartbeat.1.0.MODE_OPERATIONAL``."""
        for c in self.constants:
            if c.name == name.native_value:
                assert isinstance(c.value, _expression.Any)
                return c.value

        if name.native_value == '_extent_':  # Experimental non-standard extension
            return _expression.Rational(self.extent)

        return super(CompositeType, self)._attribute(name)  # Hand over up the inheritance chain, this is important

    def __getitem__(self, attribute_name: str) -> Attribute:
        """
        Allows the caller to retrieve an attribute by name.
        Padding fields are not accessible via this interface because they don't have names.
        Raises :class:`KeyError` if there is no such attribute.
        """
        return self._attributes_by_name[attribute_name]

    def __str__(self) -> str:
        """Returns a string like ``uavcan.node.Heartbeat.1.0``."""
        return '%s.%d.%d' % (self.full_name, self.version.major, self.version.minor)

    def __repr__(self) -> str:
        return (
            '%s(name=%r, version=%r, fields=%r, constants=%r, alignment_requirement=%r, '
            'deprecated=%r, fixed_port_id=%r)'
        ) % (
            self.__class__.__name__,
            self.full_name,
            self.version,
            self.fields,
            self.constants,
            self.alignment_requirement,
            self.deprecated,
            self.fixed_port_id,
        )


class UnionType(CompositeType):
    """
    A message type that is marked ``@union``.
    """

    MIN_NUMBER_OF_VARIANTS = 2

    def __init__(self,
                 name:             str,
                 version:          Version,
                 attributes:       typing.Iterable[Attribute],
                 deprecated:       bool,
                 fixed_port_id:    typing.Optional[int],
                 source_file_path: str,
                 parent_service:   typing.Optional['ServiceType'] = None):
        # Proxy all parameters directly to the base type - I wish we could do that
        # with kwargs while preserving the type information
        super(UnionType, self).__init__(name=name,
                                        version=version,
                                        attributes=attributes,
                                        deprecated=deprecated,
                                        fixed_port_id=fixed_port_id,
                                        source_file_path=source_file_path,
                                        parent_service=parent_service)

        if self.number_of_variants < self.MIN_NUMBER_OF_VARIANTS:
            raise MalformedUnionError('A tagged union cannot contain fewer than %d variants' %
                                      self.MIN_NUMBER_OF_VARIANTS)

        for a in attributes:
            if isinstance(a, PaddingField) or not a.name or isinstance(a.data_type, VoidType):
                raise MalformedUnionError('Padding fields not allowed in unions')

        self._tag_field_type = UnsignedIntegerType(self._compute_tag_bit_length([x.data_type for x in self.fields]),
                                                   PrimitiveType.CastMode.TRUNCATED)

    @property
    def bit_length_set(self) -> BitLengthSet:
        # Can't use @cached_property because it is unavailable before Python 3.8 and it breaks Sphinx and MyPy.
        att = '_8579621435'
        if not hasattr(self, att):
            agr = self.aggregate_bit_length_sets
            setattr(self, att, agr([f.data_type for f in self.fields]).pad_to_alignment(self.alignment_requirement))
        out = getattr(self, att)
        assert isinstance(out, BitLengthSet)
        return out

    @property
    def number_of_variants(self) -> int:
        return len(self.fields)

    @property
    def tag_field_type(self) -> UnsignedIntegerType:
        """
        The unsigned integer type of the implicit union tag field.
        Note that the set of valid tag values is a subset of that of the returned type.
        """
        return self._tag_field_type

    def iterate_fields_with_offsets(self, base_offset: typing.Optional[BitLengthSet] = None) \
            -> typing.Iterator[typing.Tuple[Field, BitLengthSet]]:
        """See the base class."""
        base_offset = BitLengthSet(base_offset or {0}).pad_to_alignment(self.alignment_requirement)
        base_offset += self.tag_field_type.bit_length
        for f in self.fields:  # Same offset for every field, because it's a tagged union, not a struct
            assert base_offset.is_aligned_at(f.data_type.alignment_requirement)
            yield f, BitLengthSet(base_offset)      # We yield a copy of the offset to prevent mutation

    @staticmethod
    def aggregate_bit_length_sets(field_types: typing.Sequence[SerializableType]) -> BitLengthSet:
        """
        Computes the bit length set for a tagged union type given the type of each of its variants.
        The final padding is not applied.

        Unions are easy to handle because when serialized, a union is essentially just a single field prefixed with
        a fixed-length integer tag. So we just build a full set of combinations and then add the tag length
        to each element.

        Observe that unions are not defined for less than 2 elements;
        however, this function tries to be generic by properly handling those cases as well,
        even though they are not permitted by the specification.
        For zero fields, the function yields ``{0}``; for one field, the function yields the BLS of the field itself.
        """
        ms = [x.bit_length_set for x in field_types]
        if len(ms) == 0:
            return BitLengthSet(0)
        if len(ms) == 1:
            return BitLengthSet(ms[0])

        tbl = UnionType._compute_tag_bit_length(field_types)
        out = BitLengthSet()
        for s in ms:
            out |= s + tbl
        assert len(out) > 0, 'Empty sets forbidden'
        return out

    @staticmethod
    def _compute_tag_bit_length(field_types: typing.Sequence[SerializableType]) -> int:
        assert len(field_types) > 1, 'Internal API misuse'
        unaligned_tag_bit_length = (len(field_types) - 1).bit_length()
        tag_bit_length = 2 ** math.ceil(math.log2(max(SerializableType.BITS_PER_BYTE, unaligned_tag_bit_length)))
        # This is to prevent the tag from breaking the alignment of the following variant.
        tag_bit_length = max([tag_bit_length] + [x.alignment_requirement for x in field_types])
        assert isinstance(tag_bit_length, int)
        assert tag_bit_length in {8, 16, 32, 64}
        return tag_bit_length


class StructureType(CompositeType):
    """
    A message type that is NOT marked ``@union``.
    """

    def iterate_fields_with_offsets(self, base_offset: typing.Optional[BitLengthSet] = None) \
            -> typing.Iterator[typing.Tuple[Field, BitLengthSet]]:
        """See the base class."""
        base_offset = BitLengthSet(base_offset or 0).pad_to_alignment(self.alignment_requirement)
        for f in self.fields:
            base_offset = base_offset.pad_to_alignment(f.data_type.alignment_requirement)
            yield f, BitLengthSet(base_offset)      # We yield a copy of the offset to prevent mutation
            base_offset += f.data_type.bit_length_set

    @property
    def bit_length_set(self) -> BitLengthSet:
        # Can't use @cached_property because it is unavailable before Python 3.8 and it breaks Sphinx and MyPy.
        att = '_7953874601'
        if not hasattr(self, att):
            agr = self.aggregate_bit_length_sets
            setattr(self, att, agr([f.data_type for f in self.fields]).pad_to_alignment(self.alignment_requirement))
        out = getattr(self, att)
        assert isinstance(out, BitLengthSet)
        return out

    @staticmethod
    def aggregate_bit_length_sets(field_types: typing.Sequence[SerializableType]) -> BitLengthSet:
        """
        Computes the bit length set for a structure type given the type of each of its fields.
        The final padding is not applied.
        """
        bls = BitLengthSet()
        for t in field_types:
            bls = bls.pad_to_alignment(t.alignment_requirement)
            bls += t.bit_length_set
        return bls or BitLengthSet(0)  # Empty bit length sets are forbidden


class DelimitedType(CompositeType):
    """
    Composites that are not sealed are wrapped into this container.
    It is a decorator over a composite type instance that injects the extent, bit length set, and field iteration
    logic that is specific to delimited (appendable, non-sealed) types.

    Most of the attributes are copied from the wrapped type (e.g., name, fixed port-ID, attributes, etc.),
    except for those that relate to the bit layout.

    Non-sealed composites are serialized into delimited opaque containers like ``uint8[<=(extent + 7) // 8]``,
    where the implicit length prefix is of type :attr:`delimiter_header_type`.
    Their bit length set is also computed as if it was an array as declared above,
    in order to prevent the containing definitions from making assumptions about the offsets of the following fields
    that might not survive the evolution of the type (e.g., version 1 may be 64 bits long, version 2 might be
    56 bits long, then version 3 could grow to 96 bits, unpredictable).
    """

    DEFAULT_EXTENT_MULTIPLIER = fractions.Fraction(3, 2)
    """
    If the extent is not specified explicitly, it is computed by multiplying the extent of the inner type by this.
    """

    _DEFAULT_DELIMITER_HEADER_BIT_LENGTH = 32

    def __init__(self, inner: CompositeType, extent: typing.Optional[int]):
        self._inner = inner
        super(DelimitedType, self).__init__(name=inner.full_name,
                                            version=inner.version,
                                            attributes=inner.attributes,
                                            deprecated=inner.deprecated,
                                            fixed_port_id=inner.fixed_port_id,
                                            source_file_path=inner.source_file_path,
                                            parent_service=inner.parent_service)
        if extent is None:
            unaligned = math.floor(inner.extent * self.DEFAULT_EXTENT_MULTIPLIER)
            self._extent = max(BitLengthSet(unaligned).pad_to_alignment(self.alignment_requirement))
        else:
            self._extent = int(extent)

        if self._extent % self.alignment_requirement != 0:
            raise InvalidExtentError('The specified extent of %d bits is not a multiple of %d bits' %
                                     (self._extent, self.alignment_requirement))
        if self._extent < inner.extent:
            raise InvalidExtentError(
                'The specified extent of %d bits is too small for this data type. '
                'Either compactify the data type or increase the extent at least to %d bits. '
                'Beware that the latter option may break wire compatibility.' %
                (self._extent, inner.extent)
            )

        # Invariant checks.
        assert self.extent % self.BITS_PER_BYTE == 0
        assert self.extent % self.alignment_requirement == 0
        assert self.extent >= self.inner_type.extent
        assert len(self.bit_length_set) > 0
        assert self.bit_length_set.is_aligned_at_byte()
        assert self.bit_length_set.is_aligned_at(self.alignment_requirement)
        assert not self.bit_length_set or \
            self.extent >= max(self.bit_length_set) - self.delimiter_header_type.bit_length

    @property
    def inner_type(self) -> CompositeType:
        """
        The appendable type that is serialized inside this delimited container.
        Its bit length set, extent, and other layout-specific entities are computed as if it was a sealed type.
        """
        return self._inner

    @property
    def extent(self) -> int:
        """
        The extent of a delimited type can be specified explicitly via ``@extent`` (provided that it is not less
        than the minimum); otherwise, it defaults to ``floor(inner_type.extent * 3/2)`` padded to byte.

        Optional optimization hint: if the objective is to allocate buffer memory for constructing a new
        serialized representation locally, then it may be beneficial to use the extent of the inner type
        rather than this one because it may be smaller. This is not safe for deserialization, of course.
        """
        return self._extent

    @property
    def bit_length_set(self) -> BitLengthSet:
        """
        For a non-sealed type, not many guarantees about the bit length set can be provided,
        because the type may be mutated in the next minor revision.
        Therefore, a synthetic bit length set is constructed that is merely a list of all possible bit lengths
        plus the delimiter header.
        For example, a type that contains a single field of type ``uint32[2]`` would have the bit length set of
        ``{h, h+8, h+16, ..., h+56, h+64}`` where ``h`` is the length of the delimiter header.
        """
        # Can't use @cached_property because it is unavailable before Python 3.8 and it breaks Sphinx and MyPy.
        att = '_3476583631'
        if not hasattr(self, att):
            x = BitLengthSet(range(self.extent + 1)).pad_to_alignment(self.alignment_requirement) + \
                self.delimiter_header_type.bit_length
            setattr(self, att, x)
        out = getattr(self, att)
        assert isinstance(out, BitLengthSet)
        return out

    @property
    def delimiter_header_type(self) -> UnsignedIntegerType:
        """
        The type of the integer prefix field that encodes the size of the serialized representation [in bytes]
        of the :attr:`inner_type`.
        """
        bit_length = self._DEFAULT_DELIMITER_HEADER_BIT_LENGTH  # This may be made configurable later.
        # This is to prevent the delimiter header from breaking the alignment of the following composite.
        bit_length = max(bit_length, self.alignment_requirement)
        return UnsignedIntegerType(bit_length, UnsignedIntegerType.CastMode.SATURATED)

    def iterate_fields_with_offsets(self, base_offset: typing.Optional[BitLengthSet] = None) \
            -> typing.Iterator[typing.Tuple[Field, BitLengthSet]]:
        """
        Delegates the call to the inner type, but with the base offset increased by the size of the delimiter header.
        """
        base_offset = (base_offset or BitLengthSet(0)) + self.delimiter_header_type.bit_length_set
        return self.inner_type.iterate_fields_with_offsets(base_offset)

    def __repr__(self) -> str:
        return '%s(inner=%r, extent=%r)' % (self.__class__.__name__, self.inner_type, self.extent)


class ServiceType(CompositeType):
    """
    A service (not message) type.
    Unlike message types, it can't be serialized directly.

    There are exactly two pseudo-fields: ``request`` and ``response``,
    which contain the request and the response structure of the service type, respectively.
    """

    class SchemaParams:
        """A trivial helper dataclass used for constructing new instances."""
        def __init__(self,
                     attributes: typing.Iterable[Attribute],
                     extent:     typing.Optional[int],
                     is_sealed:  bool,
                     is_union:   bool):
            self.attributes = list(attributes)
            self.extent = int(extent) if extent is not None else None
            self.is_sealed = bool(is_sealed)
            self.is_union = bool(is_union)
            if self.is_sealed and self.extent is not None:  # pragma: no cover
                raise ValueError('API misuse: cannot set the extent on a sealed type')

        def construct_composite(self,
                                name: str,
                                version: Version,
                                deprecated: bool,
                                parent_service: 'ServiceType') -> CompositeType:
            request_meta_type = UnionType if self.is_union else StructureType  # type: type
            ty = request_meta_type(name=name,
                                   version=version,
                                   attributes=self.attributes,
                                   deprecated=deprecated,
                                   fixed_port_id=None,
                                   source_file_path='',
                                   parent_service=parent_service)
            assert isinstance(ty, CompositeType)
            if self.is_sealed:
                assert self.extent is None
                return ty
            else:
                return DelimitedType(ty, extent=self.extent)

    def __init__(self,
                 name:             str,
                 version:          Version,
                 request_params:   SchemaParams,
                 response_params:  SchemaParams,
                 deprecated:       bool,
                 fixed_port_id:    typing.Optional[int],
                 source_file_path: str):
        self._request_type = request_params.construct_composite(name=name + '.Request',
                                                                version=version,
                                                                deprecated=deprecated,
                                                                parent_service=self)
        self._response_type = response_params.construct_composite(name=name + '.Response',
                                                                  version=version,
                                                                  deprecated=deprecated,
                                                                  parent_service=self)
        container_attributes = [
            Field(data_type=self._request_type,  name='request'),
            Field(data_type=self._response_type, name='response'),
        ]
        super(ServiceType, self).__init__(name=name,
                                          version=version,
                                          attributes=container_attributes,
                                          deprecated=deprecated,
                                          fixed_port_id=fixed_port_id,
                                          source_file_path=source_file_path)
        assert self.request_type.parent_service is self
        assert self.response_type.parent_service is self

    @property
    def bit_length_set(self) -> BitLengthSet:
        raise TypeError('Service types are not directly serializable. Use either request or response.')

    @property
    def request_type(self) -> CompositeType:
        """The type of the request schema."""
        return self._request_type

    @property
    def response_type(self) -> CompositeType:
        """The type of the response schema."""
        return self._response_type

    def iterate_fields_with_offsets(self, base_offset: typing.Optional[BitLengthSet] = None) \
            -> typing.Iterator[typing.Tuple[Field, BitLengthSet]]:
        """Always raises a :class:`TypeError`."""
        raise TypeError('Service types do not have serializable fields. Use either request or response.')


def _unittest_composite_types() -> None:
    from pytest import raises
    from ._primitive import SignedIntegerType, FloatType
    from ._array import FixedLengthArrayType, VariableLengthArrayType

    def try_name(name: str) -> CompositeType:
        return StructureType(name=name,
                             version=Version(0, 1),
                             attributes=[],
                             deprecated=False,
                             fixed_port_id=None,
                             source_file_path='')

    with raises(InvalidNameError, match='(?i).*empty.*'):
        try_name('')

    with raises(InvalidNameError, match='(?i).*root namespace.*'):
        try_name('T')

    with raises(InvalidNameError, match='(?i).*long.*'):
        try_name('namespace.another.deeper.' * 10 + 'LongTypeName')

    with raises(InvalidNameError, match='(?i).*component.*empty.*'):
        try_name('namespace.ns..T')

    with raises(InvalidNameError, match='(?i).*component.*empty.*'):
        try_name('.namespace.ns.T')

    with raises(InvalidNameError, match='(?i).*cannot start with.*'):
        try_name('namespace.0ns.T')

    with raises(InvalidNameError, match='(?i).*cannot start with.*'):
        try_name('namespace.ns.0T')

    with raises(InvalidNameError, match='(?i).*cannot contain.*'):
        try_name('namespace.n-s.T')

    assert try_name('root.nested.T').full_name == 'root.nested.T'
    assert try_name('root.nested.T').full_namespace == 'root.nested'
    assert try_name('root.nested.T').root_namespace == 'root'
    assert try_name('root.nested.T').short_name == 'T'

    print(ServiceType(name='a' * 48 + '.T',     # No exception raised
                      version=Version(0, 1),
                      request_params=ServiceType.SchemaParams([], None, False, False),
                      response_params=ServiceType.SchemaParams([], None, False, False),
                      deprecated=False,
                      fixed_port_id=None,
                      source_file_path=''))

    with raises(MalformedUnionError, match='.*variants.*'):
        UnionType(name='a.A',
                  version=Version(0, 1),
                  attributes=[],
                  deprecated=False,
                  fixed_port_id=None,
                  source_file_path='')

    with raises(MalformedUnionError, match='(?i).*padding.*'):
        UnionType(name='a.A',
                  version=Version(0, 1),
                  attributes=[
                      Field(UnsignedIntegerType(16, PrimitiveType.CastMode.TRUNCATED), 'a'),
                      Field(SignedIntegerType(16, PrimitiveType.CastMode.SATURATED), 'b'),
                      PaddingField(VoidType(16)),
                  ],
                  deprecated=False,
                  fixed_port_id=None,
                  source_file_path='')

    u = UnionType(name='uavcan.node.Heartbeat',
                  version=Version(42, 123),
                  attributes=[
                      Field(UnsignedIntegerType(16, PrimitiveType.CastMode.TRUNCATED), 'a'),
                      Field(SignedIntegerType(16, PrimitiveType.CastMode.SATURATED), 'b'),
                      Constant(FloatType(32, PrimitiveType.CastMode.SATURATED), 'A', _expression.Rational(123)),
                  ],
                  deprecated=False,
                  fixed_port_id=None,
                  source_file_path='')
    assert u['a'].name == 'a'
    assert u['b'].name == 'b'
    assert u['A'].name == 'A'
    assert u.fields == u.fields_except_padding
    with raises(KeyError):
        assert u['c']
    assert hash(u) == hash(u)
    del u

    s = StructureType(name='a.A',
                      version=Version(0, 1),
                      attributes=[
                          PaddingField(VoidType(8)),
                          Field(UnsignedIntegerType(16, PrimitiveType.CastMode.TRUNCATED), 'a'),
                          PaddingField(VoidType(64)),
                          Field(SignedIntegerType(16, PrimitiveType.CastMode.SATURATED), 'b'),
                          PaddingField(VoidType(2)),
                          Constant(FloatType(32, PrimitiveType.CastMode.SATURATED), 'A', _expression.Rational(123)),
                      ],
                      deprecated=False,
                      fixed_port_id=None,
                      source_file_path='')
    assert s['a'].name == 'a'
    assert s['b'].name == 'b'
    assert s['A'].name == 'A'
    assert len(s.constants) == 1
    assert len(s.fields) == 5
    assert len(s.fields_except_padding) == 2
    with raises(KeyError):
        assert s['c']
    with raises(KeyError):
        assert s['']        # Padding fields are not accessible
    assert hash(s) == hash(s)

    d = DelimitedType(s, None)
    assert d.inner_type is s
    assert d.attributes == d.inner_type.attributes
    with raises(KeyError):
        assert d['c']
    assert hash(d) == hash(d)
    assert d.delimiter_header_type.bit_length == 32
    assert isinstance(d.delimiter_header_type, UnsignedIntegerType)
    assert d.extent == d.inner_type.extent * 3 // 2

    d = DelimitedType(s, 256)
    assert hash(d) == hash(d)
    assert d.delimiter_header_type.bit_length == 32
    assert isinstance(d.delimiter_header_type, UnsignedIntegerType)
    assert d.extent == 256

    d = DelimitedType(s, s.extent)  # Minimal extent
    assert hash(d) == hash(d)
    assert d.delimiter_header_type.bit_length == 32
    assert isinstance(d.delimiter_header_type, UnsignedIntegerType)
    assert d.extent == s.extent == d.inner_type.extent

    with raises(InvalidExtentError):
        assert DelimitedType(s, 255)  # Unaligned extent

    with raises(InvalidExtentError):
        assert DelimitedType(s, 8)  # Extent too small

    def try_union_fields(field_types: typing.List[SerializableType]) -> UnionType:
        atr = []
        for i, t in enumerate(field_types):
            atr.append(Field(t, '_%d' % i))

        return UnionType(name='a.A',
                         version=Version(0, 1),
                         attributes=atr,
                         deprecated=False,
                         fixed_port_id=None,
                         source_file_path='')

    u = try_union_fields([
        UnsignedIntegerType(16, PrimitiveType.CastMode.TRUNCATED),
        SignedIntegerType(16, PrimitiveType.CastMode.SATURATED),
    ])
    assert u.bit_length_set == {24}
    assert u.extent == 24
    assert DelimitedType(u, None).extent == 40
    assert DelimitedType(u, None).bit_length_set == {32, 40, 48, 56, 64, 72}
    assert DelimitedType(u, 24).extent == 24
    assert DelimitedType(u, 24).bit_length_set == {32, 40, 48, 56}
    assert DelimitedType(u, 32).extent == 32
    assert DelimitedType(u, 32).bit_length_set == {32, 40, 48, 56, 64}
    assert DelimitedType(u, 800).extent == 800

    assert try_union_fields(
        [
            UnsignedIntegerType(16, PrimitiveType.CastMode.TRUNCATED),
            SignedIntegerType(16, PrimitiveType.CastMode.SATURATED),
        ] * 257
    ).bit_length_set == {16 + 16}

    assert try_union_fields(
        [
            UnsignedIntegerType(16, PrimitiveType.CastMode.TRUNCATED),
            SignedIntegerType(16, PrimitiveType.CastMode.SATURATED),
        ] * 32769
    ).bit_length_set == {32 + 16}

    # The reference values for the following test are explained in the array tests above
    tu8 = UnsignedIntegerType(8, cast_mode=PrimitiveType.CastMode.TRUNCATED)
    small = VariableLengthArrayType(tu8, 2)
    outer = FixedLengthArrayType(small, 2)   # unpadded bit length values: {4, 12, 20, 28, 36}

    # Above plus one bit to each, plus 16-bit for the unsigned integer field
    assert try_union_fields([
        outer,
        SignedIntegerType(16, PrimitiveType.CastMode.SATURATED),
    ]).bit_length_set == {24, 32, 40, 48, 56}

    def try_struct_fields(field_types: typing.List[SerializableType]) -> StructureType:
        atr = []
        for i, t in enumerate(field_types):
            atr.append(Field(t, '_%d' % i))

        return StructureType(name='a.A',
                             version=Version(0, 1),
                             attributes=atr,
                             deprecated=False,
                             fixed_port_id=None,
                             source_file_path='')

    s = try_struct_fields([
        UnsignedIntegerType(16, PrimitiveType.CastMode.TRUNCATED),
        SignedIntegerType(16, PrimitiveType.CastMode.SATURATED),
    ])
    assert s.bit_length_set == {32}
    assert s.extent == 32
    assert DelimitedType(s, None).extent == 48
    assert DelimitedType(s, None).bit_length_set == {32, 40, 48, 56, 64, 72, 80}
    assert DelimitedType(s, 32).extent == 32
    assert DelimitedType(s, 32).bit_length_set == {32, 40, 48, 56, 64}
    assert DelimitedType(s, 40).extent == 40
    assert DelimitedType(s, 40).bit_length_set == {32, 40, 48, 56, 64, 72}

    assert try_struct_fields([]).bit_length_set == {0}   # Empty sets forbidden

    assert try_struct_fields([
        outer,
        SignedIntegerType(16, PrimitiveType.CastMode.SATURATED),
    ]).bit_length_set == {16 + 16, 24 + 16, 32 + 16, 40 + 16, 48 + 16}

    assert try_struct_fields([outer]).bit_length_set == {16, 24, 32, 40, 48}


def _unittest_field_iterators() -> None:
    from pytest import raises
    from ._primitive import BooleanType, FloatType
    from ._array import FixedLengthArrayType, VariableLengthArrayType

    saturated = PrimitiveType.CastMode.SATURATED
    _seq_no = 0

    def make_type(meta: typing.Type[CompositeType], attributes: typing.Iterable[Attribute]) -> CompositeType:
        nonlocal _seq_no
        _seq_no += 1
        return meta('ns.Type' + str(_seq_no),
                    version=Version(1, 0),
                    attributes=attributes,
                    deprecated=False,
                    fixed_port_id=None,
                    source_file_path='')

    def validate_iterator(t: CompositeType,
                          reference: typing.Iterable[typing.Tuple[str, typing.Set[int]]],
                          base_offset: typing.Optional[BitLengthSet] = None) -> None:
        for (name, ref_set), (field, real_set) in itertools.zip_longest(reference,
                                                                        t.iterate_fields_with_offsets(base_offset)):
            assert field.name == name
            assert real_set == ref_set, field.name + ': ' + str(real_set)

    a = make_type(StructureType, [
        Field(UnsignedIntegerType(10, saturated), 'a'),
        Field(BooleanType(saturated), 'b'),
        Field(VariableLengthArrayType(FloatType(32, saturated), 2), 'c'),
        Field(FixedLengthArrayType(FloatType(32, saturated), 7), 'd'),
        PaddingField(VoidType(3)),
    ])

    validate_iterator(a, [
        ('a', {0}),
        ('b', {10}),
        ('c', {11}),
        ('d', {
            11 + 8 + 32 * 0,
            11 + 8 + 32 * 1,
            11 + 8 + 32 * 2,
        }),
        ('', {
            11 + 8 + 32 * 0 + 32 * 7,
            11 + 8 + 32 * 1 + 32 * 7,
            11 + 8 + 32 * 2 + 32 * 7,
        }),
    ])

    d = DelimitedType(a, None)
    validate_iterator(d, [
        ('a', {32 + 0}),
        ('b', {32 + 10}),
        ('c', {32 + 11}),
        ('d', {
            32 + 11 + 8 + 32 * 0,
            32 + 11 + 8 + 32 * 1,
            32 + 11 + 8 + 32 * 2,
        }),
        ('', {
            32 + 11 + 8 + 32 * 0 + 32 * 7,
            32 + 11 + 8 + 32 * 1 + 32 * 7,
            32 + 11 + 8 + 32 * 2 + 32 * 7,
        }),
    ])
    print('d.bit_length_set', d.bit_length_set)
    assert d.bit_length_set == BitLengthSet({
        32 + x for x in range(((11 + 8 + 32 * 2 + 32 * 7) + 7) // 8 * 8 * 3 // 2 + 1)
    }).pad_to_alignment(8)

    a_bls_options = [
        11 + 8 + 32 * 0 + 32 * 7 + 3,
        11 + 8 + 32 * 1 + 32 * 7 + 3,
        11 + 8 + 32 * 2 + 32 * 7 + 3,
    ]
    assert a.bit_length_set == BitLengthSet(a_bls_options).pad_to_alignment(8)

    # Testing "a" again, this time with non-zero base offset.
    # The first base offset element is one, but it is padded to byte, so it becomes 8.
    validate_iterator(a, [
        ('a', {8, 16}),
        ('b', {8 + 10, 16 + 10}),
        ('c', {8 + 11, 16 + 11}),
        ('d', {
            8 + 11 + 8 + 32 * 0,
            8 + 11 + 8 + 32 * 1,
            8 + 11 + 8 + 32 * 2,
            16 + 11 + 8 + 32 * 0,
            16 + 11 + 8 + 32 * 1,
            16 + 11 + 8 + 32 * 2,
        }),
        ('', {
            8 + 11 + 8 + 32 * 0 + 32 * 7,
            8 + 11 + 8 + 32 * 1 + 32 * 7,
            8 + 11 + 8 + 32 * 2 + 32 * 7,
            16 + 11 + 8 + 32 * 0 + 32 * 7,
            16 + 11 + 8 + 32 * 1 + 32 * 7,
            16 + 11 + 8 + 32 * 2 + 32 * 7,
        }),
    ], BitLengthSet({1, 16}))  # 1 becomes 8 due to padding.

    # Wrap the above into a delimited type with a manually specified extent.
    d = DelimitedType(a, 400)
    validate_iterator(d, [
        ('a', {32 + 8, 32 + 16}),
        ('b', {32 + 8 + 10, 32 + 16 + 10}),
        ('c', {32 + 8 + 11, 32 + 16 + 11}),
        ('d', {
            32 + 8 + 11 + 8 + 32 * 0,
            32 + 8 + 11 + 8 + 32 * 1,
            32 + 8 + 11 + 8 + 32 * 2,
            32 + 16 + 11 + 8 + 32 * 0,
            32 + 16 + 11 + 8 + 32 * 1,
            32 + 16 + 11 + 8 + 32 * 2,
        }),
        ('', {
            32 + 8 + 11 + 8 + 32 * 0 + 32 * 7,
            32 + 8 + 11 + 8 + 32 * 1 + 32 * 7,
            32 + 8 + 11 + 8 + 32 * 2 + 32 * 7,
            32 + 16 + 11 + 8 + 32 * 0 + 32 * 7,
            32 + 16 + 11 + 8 + 32 * 1 + 32 * 7,
            32 + 16 + 11 + 8 + 32 * 2 + 32 * 7,
        }),
    ], BitLengthSet({1, 16}))  # 1 becomes 8 due to padding.
    assert d.bit_length_set == BitLengthSet({(32 + x + 7) // 8 * 8 for x in range(400 + 1)})

    b = make_type(StructureType, [
        Field(a, 'z'),
        Field(VariableLengthArrayType(a, 2), 'y'),
        Field(UnsignedIntegerType(6, saturated), 'x'),
    ])

    a_bls_padded = [((x + 7) // 8) * 8 for x in a_bls_options]
    validate_iterator(b, [
        ('z', {0}),
        ('y', {
            a_bls_padded[0],
            a_bls_padded[1],
            a_bls_padded[2],
        }),
        ('x', {  # The lone "+2" is for the variable-length array's implicit length field
            # First length option of z
            a_bls_padded[0] + 8 + a_bls_padded[0] * 0,  # suka
            a_bls_padded[0] + 8 + a_bls_padded[1] * 0,
            a_bls_padded[0] + 8 + a_bls_padded[2] * 0,
            a_bls_padded[0] + 8 + a_bls_padded[0] * 1,
            a_bls_padded[0] + 8 + a_bls_padded[1] * 1,
            a_bls_padded[0] + 8 + a_bls_padded[2] * 1,
            a_bls_padded[0] + 8 + a_bls_padded[0] * 2,
            a_bls_padded[0] + 8 + a_bls_padded[1] * 2,
            a_bls_padded[0] + 8 + a_bls_padded[2] * 2,
            # Second length option of z
            a_bls_padded[1] + 8 + a_bls_padded[0] * 0,
            a_bls_padded[1] + 8 + a_bls_padded[1] * 0,
            a_bls_padded[1] + 8 + a_bls_padded[2] * 0,
            a_bls_padded[1] + 8 + a_bls_padded[0] * 1,
            a_bls_padded[1] + 8 + a_bls_padded[1] * 1,
            a_bls_padded[1] + 8 + a_bls_padded[2] * 1,
            a_bls_padded[1] + 8 + a_bls_padded[0] * 2,
            a_bls_padded[1] + 8 + a_bls_padded[1] * 2,
            a_bls_padded[1] + 8 + a_bls_padded[2] * 2,
            # Third length option of z
            a_bls_padded[2] + 8 + a_bls_padded[0] * 0,
            a_bls_padded[2] + 8 + a_bls_padded[1] * 0,
            a_bls_padded[2] + 8 + a_bls_padded[2] * 0,
            a_bls_padded[2] + 8 + a_bls_padded[0] * 1,
            a_bls_padded[2] + 8 + a_bls_padded[1] * 1,
            a_bls_padded[2] + 8 + a_bls_padded[2] * 1,
            a_bls_padded[2] + 8 + a_bls_padded[0] * 2,
            a_bls_padded[2] + 8 + a_bls_padded[1] * 2,
            a_bls_padded[2] + 8 + a_bls_padded[2] * 2,
        }),
    ])

    # Ensuring the equivalency between bit length and aligned bit offset
    b_offset = BitLengthSet()
    for f in b.fields:
        b_offset += f.data_type.bit_length_set
    print('b_offset:', b_offset)
    assert b_offset.pad_to_alignment(8) == b.bit_length_set
    assert not b_offset.is_aligned_at_byte()
    assert not b_offset.is_aligned_at(32)

    c = make_type(UnionType, [
        Field(a, 'foo'),
        Field(b, 'bar'),
    ])

    validate_iterator(c, [
        ('foo', {8}),       # The offset is the same because it's a union
        ('bar', {8}),
    ])

    validate_iterator(c, [
        ('foo', {8 + 8}),
        ('bar', {8 + 8}),
    ], BitLengthSet(8))

    validate_iterator(c, [
        ('foo', {0 + 8, 8 + 8}),
        ('bar', {0 + 8, 8 + 8}),
    ], BitLengthSet({0, 4, 8}))  # The option 4 is eliminated due to padding to byte, so we're left with {0, 8}.

    with raises(TypeError, match='.*request or response.*'):
        ServiceType(name='ns.S',
                    version=Version(1, 0),
                    request_params=ServiceType.SchemaParams([], None, False, False),
                    response_params=ServiceType.SchemaParams([], None, False, False),
                    deprecated=False,
                    fixed_port_id=None,
                    source_file_path='').iterate_fields_with_offsets()

    # Check the auto-padding logic.
    e = StructureType(name='e.E',
                      version=Version(0, 1),
                      attributes=[],
                      deprecated=False,
                      fixed_port_id=None,
                      source_file_path='')
    validate_iterator(e, [])
    a = make_type(StructureType, [
        Field(UnsignedIntegerType(3, PrimitiveType.CastMode.TRUNCATED), 'x'),
        Field(e, 'y'),
        Field(UnsignedIntegerType(2, PrimitiveType.CastMode.TRUNCATED), 'z'),
    ])
    assert a.bit_length_set == {16}
    validate_iterator(a, [
        ('x', {0}),
        ('y', {8}),  # Padded out!
        ('z', {8}),
    ])
